import torch.nn as nn

ACTIVATIONS = {
    'relu': nn.ReLU(),
    'leaky_relu': nn.LeakyReLU(),
    'linear': nn.Identity(),
    'sigmoid': nn.Sigmoid(),
    'softmax': nn.Softmax(),
    'log_softmax': nn.LogSoftmax(),
    'elu': nn.ELU(),
}


import torch
import numpy as np
import logging

logger = logging.getLogger(__name__)


def one_hot_encoder(idx, n_cls):
    assert torch.max(idx).item() < n_cls
    if idx.dim() == 1:
        idx = idx.unsqueeze(1)
    onehot = torch.zeros(idx.size(0), n_cls)
    onehot = onehot.to(idx.device)
    onehot.scatter_(1, idx.long(), 1)
    return onehot


def partition(data, partitions, num_partitions):
    res = []
    partitions = partitions.flatten()
    for i in range(num_partitions):
        indices = torch.nonzero((partitions == i), as_tuple=False).squeeze(1)
        res += [data[indices]]
    return res


def _validate_var_names(adata, source_var_names):
    user_var_names = adata.var_names.astype(str)
    if not np.array_equal(source_var_names, user_var_names):
        logger.warning(
            "var_names for adata passed in does not match var_names of "
            "adata used to train the model. For valid results, the vars "
            "need to be the same and in the same order as the adata used to train the model."
        )



from .trvae import AnnotatedDataset as trVAEDataset, label_encoder, remove_sparsity
from scvi.data import setup_anndata

__all__ = ('label_encoder', 'remove_sparsity', 'setup_anndata', 'trVAEDataset')



import torch
import numpy as np
from functools import partial
from torch.autograd import Variable
import torch.nn.functional as F

from ._utils import partition


def mse(recon_x, x):
    """Computes MSE loss between reconstructed data and ground truth data.

       Parameters
       ----------
       recon_x: torch.Tensor
            Torch Tensor of reconstructed data
       x: torch.Tensor
            Torch Tensor of ground truth data

       Returns
       -------
       MSE loss value
    """
    mse_loss = torch.nn.functional.mse_loss(recon_x, x, reduction='none')
    return mse_loss


def nb(x: torch.Tensor, mu: torch.Tensor, theta: torch.Tensor, eps=1e-8):
    """Computes negative binomial loss.

       Parameters
       ----------
       x: torch.Tensor
            Torch Tensor of ground truth data.
       mu: torch.Tensor
            Torch Tensor of means of the negative binomial (has to be positive support).
       theta: torch.Tensor
            Torch Tensor of inverse dispersion parameter (has to be positive support).
       eps: Float
            numerical stability constant.

       Returns
       -------
       If 'mean' is 'True' NB loss value gets returned, otherwise Torch tensor of losses gets returned.
    """
    if theta.ndimension() == 1:
        theta = theta.view(1, theta.size(0))

    log_theta_mu_eps = torch.log(theta + mu + eps)
    res = (
        theta * (torch.log(theta + eps) - log_theta_mu_eps)
        + x * (torch.log(mu + eps) - log_theta_mu_eps)
        + torch.lgamma(x + theta)
        - torch.lgamma(theta)
        - torch.lgamma(x + 1)
    )

    return res


def zinb(x: torch.Tensor, mu: torch.Tensor, theta: torch.Tensor, pi: torch.Tensor, eps=1e-8):
    """Computes zero inflated negative binomial loss.

       Parameters
       ----------
       x: torch.Tensor
            Torch Tensor of ground truth data.
       mu: torch.Tensor
            Torch Tensor of means of the negative binomial (has to be positive support).
       theta: torch.Tensor
            Torch Tensor of inverses dispersion parameter (has to be positive support).
       pi: torch.Tensor
            Torch Tensor of logits of the dropout parameter (real support)
       eps: Float
            numerical stability constant.

       Returns
       -------
       If 'mean' is 'True' ZINB loss value gets returned, otherwise Torch tensor of losses gets returned.
    """
    # theta is the dispersion rate. If .ndimension() == 1, it is shared for all cells (regardless of batch or labels)
    if theta.ndimension() == 1:
        theta = theta.view(
            1, theta.size(0)
        )  # In this case, we reshape theta for broadcasting

    softplus_pi = F.softplus(-pi)  #  uses log(sigmoid(x)) = -softplus(-x)
    log_theta_eps = torch.log(theta + eps)
    log_theta_mu_eps = torch.log(theta + mu + eps)
    pi_theta_log = -pi + theta * (log_theta_eps - log_theta_mu_eps)

    case_zero = F.softplus(pi_theta_log) - softplus_pi
    mul_case_zero = torch.mul((x < eps).type(torch.float32), case_zero)

    case_non_zero = (
        -softplus_pi
        + pi_theta_log
        + x * (torch.log(mu + eps) - log_theta_mu_eps)
        + torch.lgamma(x + theta)
        - torch.lgamma(theta)
        - torch.lgamma(x + 1)
    )
    mul_case_non_zero = torch.mul((x > eps).type(torch.float32), case_non_zero)

    res = mul_case_zero + mul_case_non_zero
    return res


def pairwise_distance(x, y):

    if not len(x.shape) == len(y.shape) == 2:
        raise ValueError('Both inputs should be matrices.')

    if x.shape[1] != y.shape[1]:
        raise ValueError('The number of features should be the same.')

    x = x.view(x.shape[0], x.shape[1], 1)
    y = torch.transpose(y, 0, 1)
    output = torch.sum((x - y) ** 2, 1)
    output = torch.transpose(output, 0, 1)

    return output


def gaussian_kernel_matrix(x, y, sigmas):
    """Computes multiscale-RBF kernel between x and y.

       Parameters
       ----------
       x: torch.Tensor
            Tensor with shape [batch_size, z_dim].
       y: torch.Tensor
            Tensor with shape [batch_size, z_dim].
       sigmas: Tensor

       Returns
       -------
       Returns the computed multiscale-RBF kernel between x and y.
    """

    sigmas = sigmas.view(sigmas.shape[0], 1)
    beta = 1. / (2. * sigmas)
    dist = pairwise_distance(x, y).contiguous()
    dist_ = dist.view(1, -1)
    s = torch.matmul(beta, dist_)

    return torch.sum(torch.exp(-s), 0).view_as(dist)


def maximum_mean_discrepancy(x, y, kernel=gaussian_kernel_matrix):
    """Computes Maximum Mean Discrepancy(MMD) between x and y.

       Parameters
       ----------
       x: torch.Tensor
            Tensor with shape [batch_size, z_dim]
       y: torch.Tensor
            Tensor with shape [batch_size, z_dim]
       kernel: gaussian_kernel_matrix

       Returns
       -------
       Returns the computed MMD between x and y.
    """

    cost = torch.mean(kernel(x, x))
    cost += torch.mean(kernel(y, y))
    cost -= 2 * torch.mean(kernel(x, y))

    return cost


def mmd_loss_calc(source_features, target_features):
    """Initializes Maximum Mean Discrepancy(MMD) between source_features and target_features.

       Parameters
       ----------
       source_features: torch.Tensor
            Tensor with shape [batch_size, z_dim]
       target_features: torch.Tensor
            Tensor with shape [batch_size, z_dim]

       Returns
       -------
       Returns the computed MMD between x and y.
    """

    sigmas = [
        1e-6, 1e-5, 1e-4, 1e-3, 1e-2, 1e-1, 1, 5, 10, 15, 20, 25, 30, 35, 100,
        1e3, 1e4, 1e5, 1e6
    ]
    if torch.cuda.is_available():
        gaussian_kernel = partial(
            gaussian_kernel_matrix, sigmas=Variable(torch.cuda.FloatTensor(sigmas))
        )
    else:
        gaussian_kernel = partial(
            gaussian_kernel_matrix, sigmas=Variable(torch.FloatTensor(sigmas))
        )
    loss_value = maximum_mean_discrepancy(source_features, target_features, kernel=gaussian_kernel)

    return loss_value


def mmd(n_conditions, beta, boundary):
    """Initializes Maximum Mean Discrepancy(MMD) between every different condition.

       Parameters
       ----------
       n_conditions: integer
            Number of classes (conditions) the data contain.
       beta: float
            beta coefficient for MMD loss.
       boundary: integer
            If not 'None', mmd loss is only calculated on #new conditions.
       y: torch.Tensor
            Torch Tensor of computed latent data.
       c: torch.Tensor
            Torch Tensor of condition labels.

       Returns
       -------
       Returns MMD loss.
    """
    def mmd_loss(y, c):
        # partition separates y into num_cls subsets w.r.t. their labels c
        conditions_mmd = partition(y, c, n_conditions)
        loss = 0.0
        if boundary is not None:
            for i in range(boundary):
                for j in range(boundary, n_conditions):
                    if conditions_mmd[i].size(0) < 2 or conditions_mmd[j].size(0) < 2:
                        continue
                    loss += mmd_loss_calc(conditions_mmd[i], conditions_mmd[j])
        else:
            for i in range(len(conditions_mmd)):
                if conditions_mmd[i].size(0) < 1:
                    continue
                for j in range(i):
                    if conditions_mmd[j].size(0) < 1 or i == j:
                        continue
                    loss += mmd_loss_calc(conditions_mmd[i], conditions_mmd[j])
        return beta * loss

    return mmd_loss



import inspect
import os
import torch
import pickle
import numpy as np

from anndata import AnnData, read
from typing import Optional, Union

from .trvae import trVAE
from scarches.trainers.trvae.unsupervised import trVAETrainer
from ._utils import _validate_var_names

class BaseMixin:
    def _get_user_attributes(self):
        # returns all the self attributes defined in a model class, eg, self.is_trained_
        attributes = inspect.getmembers(self, lambda a: not (inspect.isroutine(a)))
        attributes = [
            a for a in attributes if not (a[0].startswith("__") and a[0].endswith("__"))
        ]
        attributes = [a for a in attributes if not a[0].startswith("_abc_")]
        return attributes

    def _get_public_attributes(self):
        public_attributes = self._get_user_attributes()
        public_attributes = {a[0]: a[1] for a in public_attributes if a[0][-1] == "_"}
        return public_attributes

    def save(
        self,
        dir_path: str,
        overwrite: bool = False,
        save_anndata: bool = False,
        **anndata_write_kwargs,
    ):
        """Save the state of the model.

           Neither the trainer optimizer state nor the trainer history are saved.

           Parameters
           ----------
           dir_path
                Path to a directory.
           overwrite
                Overwrite existing data or not. If `False` and directory
                already exists at `dir_path`, error will be raised.
           save_anndata
                If True, also saves the anndata
           anndata_write_kwargs
                Kwargs for anndata write function
        """
        # get all the public attributes
        public_attributes = self._get_public_attributes()
        # save the model state dict and the trainer state dict only
        if not os.path.exists(dir_path) or overwrite:
            os.makedirs(dir_path, exist_ok=overwrite)
        else:
            raise ValueError(
                "{} already exists. Please provide an unexisting directory for saving.".format(
                    dir_path
                )
            )

        if save_anndata:
            self.adata.write(
                os.path.join(dir_path, "adata.h5ad"), **anndata_write_kwargs
            )

        model_save_path = os.path.join(dir_path, "model_params.pt")
        attr_save_path = os.path.join(dir_path, "attr.pkl")
        varnames_save_path = os.path.join(dir_path, "var_names.csv")

        var_names = self.adata.var_names.astype(str)
        var_names = var_names.to_numpy()
        np.savetxt(varnames_save_path, var_names, fmt="%s")

        torch.save(self.model.state_dict(), model_save_path)
        with open(attr_save_path, "wb") as f:
            pickle.dump(public_attributes, f)

    def _load_expand_params_from_dict(self, state_dict):
        load_state_dict = state_dict.copy()

        device = next(self.model.parameters()).device

        new_state_dict = self.model.state_dict()
        for key, load_ten in load_state_dict.items():
            new_ten = new_state_dict[key]
            if new_ten.size() == load_ten.size():
                continue
            # new categoricals changed size
            else:
                load_ten = load_ten.to(device)
                dim_diff = new_ten.size()[-1] - load_ten.size()[-1]
                fixed_ten = torch.cat([load_ten, new_ten[..., -dim_diff:]], dim=-1)
                load_state_dict[key] = fixed_ten

        self.model.load_state_dict(load_state_dict)

    @classmethod
    def _load_params(
        cls,
        dir_path: str,
        map_location: Optional[str] = None
    ):
        setup_dict_path = os.path.join(dir_path, "attr.pkl")
        model_path = os.path.join(dir_path, "model_params.pt")
        varnames_path = os.path.join(dir_path, "var_names.csv")

        with open(setup_dict_path, "rb") as handle:
            attr_dict = pickle.load(handle)

        model_state_dict = torch.load(model_path, map_location=map_location)

        var_names = np.genfromtxt(varnames_path, delimiter=",", dtype=str)

        return attr_dict, model_state_dict, var_names

    @classmethod
    def load(
        cls,
        dir_path: str,
        adata: Optional[AnnData] = None,
    ):
        """Instantiate a model from the saved output.

           Parameters
           ----------
           dir_path
                Path to saved outputs.
           adata
                AnnData object.
                If None, will check for and load anndata saved with the model.

           Returns
           -------
                Model with loaded state dictionaries.
        """
        adata_path = os.path.join(dir_path, "adata.h5ad")

        load_adata = adata is None

        if os.path.exists(adata_path) and load_adata:
            adata = read(adata_path)
        elif not os.path.exists(adata_path) and load_adata:
            raise ValueError("Save path contains no saved anndata and no adata was passed.")

        attr_dict, model_state_dict, var_names = cls._load_params(dir_path)

        _validate_var_names(adata, var_names)
        cls._validate_adata(adata, attr_dict)
        init_params = cls._get_init_params_from_dict(attr_dict)

        model = cls(adata, **init_params)
        model.model.load_state_dict(model_state_dict)
        model.model.eval()

        model.is_trained_ = attr_dict['is_trained_']

        return model


class TRVAE(BaseMixin):
    """Model for scArches class. This class contains the implementation of Conditional Variational Auto-encoder.

       Parameters
       ----------
       adata: : `~anndata.AnnData`
            Annotated data matrix.
       condition_key: String
            column name of conditions in `adata.obs` data frame.
       conditions: List
            List of Condition names that the used data will contain to get the right encoding when used after reloading.
       hidden_layer_sizes: List
            A list of hidden layer sizes for encoder network. Decoder network will be the reversed order.
       latent_dim: Integer
            Bottleneck layer (z)  size.
       dr_rate: Float
            Dropput rate applied to all layers, if `dr_rate`==0 no dropout will be applied.
       use_mmd: Boolean
            If 'True' an additional MMD loss will be calculated on the latent dim. 'z' or the first decoder layer 'y'.
       mmd_on: String
            Choose on which layer MMD loss will be calculated on if 'use_mmd=True': 'z' for latent dim or 'y' for first
            decoder layer.
       mmd_boundary: Integer or None
            Choose on how many conditions the MMD loss should be calculated on. If 'None' MMD will be calculated on all
            conditions.
       recon_loss: String
            Definition of Reconstruction-Loss-Method, 'mse', 'nb' or 'zinb'.
       beta: Float
            Scaling Factor for MMD loss
       use_bn: Boolean
            If `True` batch normalization will be applied to layers.
       use_ln: Boolean
            If `True` layer normalization will be applied to layers.
    """
    def __init__(
        self,
        adata: AnnData,
        condition_key: str = None,
        conditions: Optional[list] = None,
        hidden_layer_sizes: list = [256, 64],
        latent_dim: int = 10,
        dr_rate: float = 0.05,
        use_mmd: bool = True,
        mmd_on: str = 'z',
        mmd_boundary: Optional[int] = None,
        recon_loss: Optional[str] = 'nb',
        beta: float = 1,
        use_bn: bool = False,
        use_ln: bool = True,
    ):
        self.adata = adata

        self.condition_key_ = condition_key

        if conditions is None:
            if condition_key is not None:
                self.conditions_ = adata.obs[condition_key].unique().tolist()
            else:
                self.conditions_ = []
        else:
            self.conditions_ = conditions

        self.hidden_layer_sizes_ = hidden_layer_sizes
        self.latent_dim_ = latent_dim
        self.dr_rate_ = dr_rate
        self.use_mmd_ = use_mmd
        self.mmd_on_ = mmd_on
        self.mmd_boundary_ = mmd_boundary
        self.recon_loss_ = recon_loss
        self.beta_ = beta
        self.use_bn_ = use_bn
        self.use_ln_ = use_ln

        self.input_dim_ = adata.n_vars

        self.model = trVAE(
            self.input_dim_,
            self.conditions_,
            self.hidden_layer_sizes_,
            self.latent_dim_,
            self.dr_rate_,
            self.use_mmd_,
            self.mmd_on_,
            self.mmd_boundary_,
            self.recon_loss_,
            self.beta_,
            self.use_bn_,
            self.use_ln_,
        )

        self.is_trained_ = False

        self.trainer = None

    def train(
        self,
        n_epochs: int = 400,
        lr: float = 1e-3,
        eps: float = 0.01,
        **kwargs
    ):
        """Train the model.

           Parameters
           ----------
           n_epochs
                Number of epochs for training the model.
           lr
                Learning rate for training the model.
           eps
                torch.optim.Adam eps parameter
           kwargs
                kwargs for the TrVAE trainer.
        """
        self.trainer = trVAETrainer(
            self.model,
            self.adata,
            condition_key=self.condition_key_,
            **kwargs)
        self.trainer.train(n_epochs, lr, eps)
        self.is_trained_ = True

    def get_latent(
        self,
        x: Optional[np.ndarray] = None,
        c: Optional[np.ndarray] = None,
        mean: bool = False
    ):
        """Map `x` in to the latent space. This function will feed data in encoder  and return  z for each sample in
           data.

           Parameters
           ----------
           x
                Numpy nd-array to be mapped to latent space. `x` has to be in shape [n_obs, input_dim].
                If None, then `self.adata.X` is used.
           c
                `numpy nd-array` of original (unencoded) desired labels for each sample.
           mean
                return mean instead of random sample from the latent space

           Returns
           -------
                Returns array containing latent space encoding of 'x'.
        """
        device = next(self.model.parameters()).device
        if x is None and c is None:
            x = self.adata.X
            if self.conditions_ is not None:
                c = self.adata.obs[self.condition_key_]

        if c is not None:
            c = np.asarray(c)
            if not set(c).issubset(self.conditions_):
                raise ValueError("Incorrect conditions")
            labels = np.zeros(c.shape[0])
            for condition, label in self.model.condition_encoder.items():
                labels[c == condition] = label
            c = torch.tensor(labels, device=device)

        x = torch.tensor(x, device=device)

        latents = []
        indices = torch.arange(x.size(0), device=device)
        subsampled_indices = indices.split(512)
        for batch in subsampled_indices:
            latent = self.model.get_latent(x[batch,:], c[batch], mean)
            latents += [latent.cpu().detach()]

        return np.array(torch.cat(latents))

    @classmethod
    def _get_init_params_from_dict(cls, dct):
        init_params = {
            'condition_key': dct['condition_key_'],
            'conditions': dct['conditions_'],
            'hidden_layer_sizes': dct['hidden_layer_sizes_'],
            'latent_dim': dct['latent_dim_'],
            'dr_rate': dct['dr_rate_'],
            'use_mmd': dct['use_mmd_'],
            'mmd_on': dct['mmd_on_'],
            'mmd_boundary': dct['mmd_boundary_'],
            'recon_loss': dct['recon_loss_'],
            'beta': dct['beta_'],
            'use_bn': dct['use_bn_'],
            'use_ln': dct['use_ln_'],
        }

        return init_params

    @classmethod
    def _validate_adata(cls, adata, dct):
        if adata.n_vars != dct['input_dim_']:
            raise ValueError("Incorrect var dimension")

        adata_conditions = adata.obs[dct['condition_key_']].unique().tolist()
        if not set(adata_conditions).issubset(dct['conditions_']):
            raise ValueError("Incorrect conditions")

    @classmethod
    def load_query_data(
        cls,
        adata: AnnData,
        reference_model: Union[str, 'TRVAE'],
        freeze: bool = True,
        freeze_expression: bool = True,
        remove_dropout: bool = True,
    ):
        """Transfer Learning function for new data. Uses old trained model and expands it for new conditions.

           Parameters
           ----------
           adata
                Query anndata object.
           reference_model
                TRVAE model to expand or a path to TRVAE model folder.
           freeze: Boolean
                If 'True' freezes every part of the network except the first layers of encoder/decoder.
           freeze_expression: Boolean
                If 'True' freeze every weight in first layers except the condition weights.
           remove_dropout: Boolean
                If 'True' remove Dropout for Transfer Learning.

           Returns
           -------
           new_model: trVAE
                New TRVAE model to train on query data.
        """
        if isinstance(reference_model, str):
            attr_dict, model_state_dict, var_names = cls._load_params(reference_model)
            _validate_var_names(adata, var_names)
        else:
            attr_dict = reference_model._get_public_attributes()
            model_state_dict = reference_model.model.state_dict()
        init_params = cls._get_init_params_from_dict(attr_dict)

        conditions = init_params['conditions']
        condition_key = init_params['condition_key']

        new_conditions = []
        adata_conditions = adata.obs[condition_key].unique().tolist()
        # Check if new conditions are already known
        for item in adata_conditions:
            if item not in conditions:
                new_conditions.append(item)

        # Add new conditions to overall conditions
        for condition in new_conditions:
            conditions.append(condition)

        if remove_dropout:
            init_params['dr_rate'] = 0.0

        new_model = cls(adata, **init_params)
        new_model._load_expand_params_from_dict(model_state_dict)

        if freeze:
            for name, p in new_model.model.named_parameters():
                p.requires_grad = False
                if 'theta' in name:
                    p.requires_grad = True
                if freeze_expression:
                    if 'cond_L.weight' in name:
                        p.requires_grad = True
                else:
                    if "L0" in name or "N0" in name:
                        p.requires_grad = True

        return new_model



from .scvi import (
    scVITrainer,
    scANVITrainer,
    totalTrainer
)

from .trvae import (
    trVAETrainer,
)


__all__ = [
    "scVITrainer",
    "scANVITrainer",
    "totalTrainer",
    "trVAETrainer",
]



import numpy as np


def label_encoder(adata, encoder=None, condition_key='condition'):
    """Encode labels of Annotated `adata` matrix.
       Parameters
       ----------
       adata: : `~anndata.AnnData`
            Annotated data matrix.
       encoder: Dict or None
            dictionary of encoded labels. if `None`, will create one.
       condition_key: String
            column name of conditions in `adata.obs` data frame.

       Returns
       -------
       labels: `~numpy.ndarray`
            Array of encoded labels
       label_encoder: Dict
            dictionary with labels and encoded labels as key, value pairs.
    """
    unique_conditions = list(np.unique(adata.obs[condition_key]))
    if encoder is None:
        encoder = {k: v for k, v in zip(sorted(unique_conditions), np.arange(len(unique_conditions)))}

    assert set(unique_conditions).issubset(set(encoder.keys()))
    labels = np.zeros(adata.shape[0])
    for condition, label in encoder.items():
        labels[adata.obs[condition_key] == condition] = label
    return labels.reshape(-1, 1), encoder



import scanpy as sc
import numpy as np
import json
import os
import torch
import scvi as scv
import scarches as sca
from scarches.plotting import SCVI_EVAL
from scarches.dataset.trvae.data_handling import remove_sparsity

data = 'pancreas'

# ------------------------------------------Parameters that can be adjusted------------------------------------------
use_scarches_version = True
n_query_labels = 0
n_epochs_surgery = 200
freeze = True
freeze_expression = True
show_plots = False

# ---------------------------------------------------DATA PREPROCESSING-------------------------------------------------
n_epochs_vae = 500
n_epochs_scanvi = 200
early_stopping_kwargs = {
    "early_stopping_metric": "elbo",
    "save_best_state_metric": "elbo",
    "patience": 10,
    "threshold": 0,
    "reduce_lr_on_plateau": True,
    "lr_patience": 8,
    "lr_factor": 0.1,
}
early_stopping_kwargs_scanvi = {
    "early_stopping_metric": "accuracy",
    "save_best_state_metric": "accuracy",
    "on": "full_dataset",
    "patience": 10,
    "threshold": 0.001,
    "reduce_lr_on_plateau": True,
    "lr_patience": 8,
    "lr_factor": 0.1,
}

arches_params = {
    "use_layer_norm": "both",
    "use_batch_norm": "none",
    "encode_covariates": use_scarches_version,
    "dropout_rate": 0.2,
    "n_layers": 2
}

condition_key = 'study'
cell_type_key = 'cell_type'
conditions_dict = dict()
cell_type_dict = dict()
results = dict()

if use_scarches_version:
    version = 'scarches'
else:
    version = 'base'
if freeze_expression:
    surgery_version = 'surgery_only_cond'
elif freeze:
    surgery_version = 'surgery_only_first'
else:
    surgery_version = 'surgery_full_retrain'

if n_query_labels == 0:
    query_label = 'unlabelled_query'
else:
    query_label = f'labelled_query_{n_query_labels}'
epoch_label = f'epochs_{n_epochs_surgery}'
save_path = os.path.expanduser(f'~/Documents/benchmark_results/scanvi_umaps/{data}/'
                               f'{version}/{query_label}/{surgery_version}_{epoch_label}/')
if not os.path.exists(save_path):
    os.makedirs(save_path)

if data == 'pancreas':
    target_batches = ["Pancreas SS2", "Pancreas CelSeq2"]
    batch_key = "study"
    cell_type_key = "cell_type"
    adata = sc.read(os.path.expanduser(f'~/Documents/datasets/pancreas_normalized.h5ad'))
elif data == 'mouse_brain':
    target_batches = ["Tabula_muris", "Zeisel"]
    batch_key = "study"
    cell_type_key = "cell_type"
    adata = sc.read(os.path.expanduser(f'~/Documents/datasets/mouse_brain_normalized.h5ad'))

adata.X = adata.raw.X
adata = remove_sparsity(adata)
celltypes = adata.obs[cell_type_key].unique().tolist()
conditions = adata.obs[condition_key].unique().tolist()

# Save Source Adata
source_adata = adata[~adata.obs[condition_key].isin(target_batches)].copy()
scv.data.setup_anndata(source_adata, batch_key=condition_key, labels_key=cell_type_key)
source_stats = source_adata.uns["_scvi"]["summary_stats"]
print('SOURCE ADATA...\n', source_adata)
print('SOURCE STATS...\n', source_stats)

# Save Target Adata
target_adata = adata[adata.obs[condition_key].isin(target_batches)].copy()
scv.data.setup_anndata(target_adata, batch_key=condition_key, labels_key=cell_type_key)
target_stats = target_adata.uns["_scvi"]["summary_stats"]
print('\nTARGET ADATA...\n', target_adata)
print('TARGET STATS...\n', target_stats)

# Train scVI Model on Reference Data
print("\nSCVI MODEL ARCH...")
scvi = sca.models.SCVI(source_adata, use_cuda=True, **arches_params)

trainer = sca.trainers.scVITrainer(
    scvi,
    train_size=0.9,
    use_cuda=True,
    frequency=1,
    silent=False,
    early_stopping_kwargs=early_stopping_kwargs
)

trainer.train(n_epochs_vae)

# Train scANVI Model on Query Data
print('\n', 20 * '###')
print("STARTING WITH SCANVI...")
scanvi = sca.models.SCANVI(
    source_adata,
    unlabeled_category='Unknown',
    classifier_parameters={'dropout_rate': 0.2, 'n_hidden': 10, 'n_layers': 1},
    pretrained_model=scvi,
    use_cuda=True,
    **arches_params
)

trainer_scanvi = sca.trainers.scANVITrainer(
    scanvi,
    n_labelled_samples_per_class=source_stats["n_cells"],
    train_size=0.9,
    use_cuda=True,
    frequency=1,
    silent=False,
    early_stopping_kwargs=early_stopping_kwargs_scanvi
)
trainer_scanvi.train(n_epochs=n_epochs_scanvi)
scanvi_eval = SCVI_EVAL(
    scanvi,
    source_adata,
    trainer_scanvi,
    cell_type_key=cell_type_key,
    batch_key=condition_key)

results["reference_ebm"] = scanvi_eval.get_ebm()
results["reference_knn"] = scanvi_eval.get_knn_purity()
results["reference_asw_b"], results["scanvi_asw_c"] = scanvi_eval.get_asw()
results["reference_nmi"] = scanvi_eval.get_nmi()
results["reference_acc"] = scanvi_eval.get_classification_accuracy()
results["reference_f1"] = scanvi_eval.get_f1_score()
scanvi_eval.plot_history(show=show_plots, save=True, dir_path=f'{save_path}reference_history')
scanvi_eval.plot_latent(show=show_plots, save=True, dir_path=f'{save_path}reference_latent')
scanvi_eval.post_adata.write_h5ad(filename=f'{save_path}reference_latent_data.h5ad')
torch.save(scanvi.state_dict(), f'{save_path}reference_model')

new_scanvi, new_trainer, target_gene_adata = sca.scvi_operate(
    scanvi,
    target_adata,
    n_epochs=n_epochs_surgery,
    early_stopping=False,
    labels_per_class=n_query_labels,
    freeze=freeze,
    freeze_expression=freeze_expression,
)
surgery_eval = SCVI_EVAL(
    new_scanvi,
    target_gene_adata,
    new_trainer,
    cell_type_key=cell_type_key,
    batch_key=condition_key
)
results["query_ebm"] = surgery_eval.get_ebm()
results["query_knn"] = surgery_eval.get_knn_purity()
results["query_asw_b"], results["surgery_asw_c"] = surgery_eval.get_asw()
results["query_nmi"] = surgery_eval.get_nmi()
results["query_acc"] = surgery_eval.get_classification_accuracy()
results["query_f1"] = surgery_eval.get_f1_score()
surgery_eval.plot_history(show=show_plots, save=True, dir_path=f'{save_path}surgery_history')
surgery_eval.plot_latent(show=show_plots, save=True, dir_path=f'{save_path}query_latent')
surgery_eval.post_adata.write_h5ad(filename=f'{save_path}query_latent_data.h5ad')
torch.save(new_scanvi.state_dict(), f'{save_path}surgery_model')

scv.data.setup_anndata(adata, batch_key=condition_key, labels_key=cell_type_key)

# Recreate Right Batch Encoding
source_conditions = source_adata.obs[condition_key].unique().tolist()
for condition in source_conditions:
    condition_label = source_adata[source_adata.obs[condition_key] == condition].obs['_scvi_batch'].unique().tolist()[0]
    conditions_dict[condition] = condition_label
target_conditions = target_gene_adata.obs[condition_key].unique().tolist()
for condition in target_conditions:
    condition_label = target_gene_adata[
        target_gene_adata.obs[condition_key] == condition].obs['_scvi_batch'].unique().tolist()[0]
    conditions_dict[condition] = condition_label
adata_conditions = adata.obs[condition_key].copy()
new_conditions = np.zeros_like(adata_conditions)
for idx in range(len(adata_conditions)):
    new_conditions[idx] = conditions_dict[adata_conditions[idx]]
adata.obs['_scvi_batch'] = new_conditions

full_eval = SCVI_EVAL(
    new_scanvi,
    adata,
    new_trainer,
    cell_type_key=cell_type_key,
    batch_key=condition_key
)
results["full_ebm"] = full_eval.get_ebm()
results["full_knn"] = full_eval.get_knn_purity()
results["full_asw_b"], results["surgery_asw_c"] = full_eval.get_asw()
results["full_nmi"] = full_eval.get_nmi()
results["full_acc"] = full_eval.get_classification_accuracy()
results["full_f1"] = full_eval.get_f1_score()
full_eval.plot_latent(show=show_plots, save=True, dir_path=f'{save_path}full_latent')
full_eval.post_adata.write_h5ad(filename=f'{save_path}full_latent_data.h5ad')

with open(f'{save_path}metrics.txt', 'w') as filehandle:
    json.dump(results, filehandle)



import numpy as np
import torch
import anndata

from scarches.models.trvae.trvae import trVAE
from scarches.trainers.trvae.unsupervised import trVAETrainer


def trvae_operate(
        network: trVAE,
        data: anndata,
        condition_key: str = None,
        size_factor_key: str = None,
        n_epochs: int = 20,
        freeze: bool = True,
        freeze_expression: bool = True,
        remove_dropout: bool = True,
) -> [trVAE, trVAETrainer]:
    """Transfer Learning function for new data. Uses old trained Network and expands it for new conditions.
       Parameters
       ----------
       network: trVAE
            A scNet model object.
       data: Anndata
            Query anndata object.
       condition_key: String
            Key where the conditions in the data can be found.
       size_factor_key: String
            Key where the size_factors in the data can be found.
       n_epochs: Integer
            Number of epochs for training the network on query data.
       freeze: Boolean
            If 'True' freezes every part of the network except the first layers of encoder/decoder.
       freeze_expression: Boolean
            If 'True' freeze every weight in first layers except the condition weights.
       remove_dropout: Boolean
            If 'True' remove Dropout for Transfer Learning.
       Returns
       -------
       new_network: trVAE
            Newly network that got trained on query data.
       new_trainer: trVAETrainer
            Trainer for the newly network.
    """
    conditions = network.conditions
    new_conditions = []
    data_conditions = data.obs[condition_key].unique().tolist()
    # Check if new conditions are already known
    for item in data_conditions:
        if item not in conditions:
            new_conditions.append(item)

    n_new_conditions = len(new_conditions)

    # Add new conditions to overall conditions
    for condition in new_conditions:
        conditions.append(condition)

    # Update DR Rate
    new_dr = network.dr_rate
    if remove_dropout:
        new_dr = 0.0

    print("Surgery to get new Network...")
    new_network = trVAE(
        network.input_dim,
        conditions=conditions,
        hidden_layer_sizes=network.hidden_layer_sizes,
        latent_dim=network.latent_dim,
        dr_rate=new_dr,
        use_mmd=network.use_mmd,
        mmd_boundary=network.mmd_boundary,
        recon_loss=network.recon_loss,
    )

    # Expand First Layer weights of encoder/decoder of old network by new conditions
    encoder_input_weights = network.encoder.FC.L0.cond_L.weight
    to_be_added_encoder_input_weights = np.random.randn(encoder_input_weights.size()[0], n_new_conditions) * np.sqrt(
        2 / (encoder_input_weights.size()[0] + 1 + encoder_input_weights.size()[1]))
    to_be_added_encoder_input_weights = torch.from_numpy(to_be_added_encoder_input_weights).float().to(network.device)
    network.encoder.FC.L0.cond_L.weight.data = torch.cat((encoder_input_weights,
                                                          to_be_added_encoder_input_weights), 1)

    decoder_input_weights = network.decoder.FirstL.L0.cond_L.weight
    to_be_added_decoder_input_weights = np.random.randn(decoder_input_weights.size()[0], n_new_conditions) * np.sqrt(
        2 / (decoder_input_weights.size()[0] + 1 + decoder_input_weights.size()[1]))
    to_be_added_decoder_input_weights = torch.from_numpy(to_be_added_decoder_input_weights).float().to(network.device)
    network.decoder.FirstL.L0.cond_L.weight.data = torch.cat((decoder_input_weights,
                                                              to_be_added_decoder_input_weights), 1)

    # Set the weights of new network to old network weights
    new_network.load_state_dict(network.state_dict())

    # Freeze parts of the network
    if freeze:
        new_network.freeze = True
        for name, p in new_network.named_parameters():
            p.requires_grad = False
            if freeze_expression:
                if 'cond_L.weight' in name:
                    p.requires_grad = True
            else:
                if "L0" in name or "B0" in name:
                    p.requires_grad = True

    new_trainer = trVAETrainer(
        new_network,
        data,
        condition_key=condition_key,
        size_factor_key=size_factor_key,
        batch_size=1024,
        n_samples=4096
    )
    new_trainer.train(
        n_epochs=n_epochs,
        lr=0.001
    )

    return new_network, new_trainer



import torch
import torch.nn as nn

from ._utils import one_hot_encoder


class CondLayers(nn.Module):
    def __init__(
            self,
            n_in: int,
            n_out: int,
            n_cond: int,
            bias: bool,
    ):
        super().__init__()
        self.n_cond = n_cond
        self.expr_L = nn.Linear(n_in, n_out, bias=bias)
        if self.n_cond != 0:
            self.cond_L = nn.Linear(self.n_cond, n_out, bias=False)

    def forward(self, x: torch.Tensor):
        if self.n_cond == 0:
            out = self.expr_L(x)
        else:
            expr, cond = torch.split(x, x.shape[1] - self.n_cond, dim=1)
            out = self.expr_L(expr) + self.cond_L(cond)
        return out


class Encoder(nn.Module):
    """ScArches Encoder class. Constructs the encoder sub-network of TRVAE and CVAE. It will transform primary space
       input to means and log. variances of latent space with n_dimensions = z_dimension.

       Parameters
       ----------
       layer_sizes: List
            List of first and hidden layer sizes
       latent_dim: Integer
            Bottleneck layer (z)  size.
       use_bn: Boolean
            If `True` batch normalization will be applied to layers.
       use_ln: Boolean
            If `True` layer normalization will be applied to layers.
       use_dr: Boolean
            If `True` dropout will applied to layers.
       dr_rate: Float
            Dropput rate applied to all layers, if `dr_rate`==0 no dropput will be applied.
       num_classes: Integer
            Number of classes (conditions) the data contain. if `None` the model will be a normal VAE instead of
            conditional VAE.
    """
    def __init__(self,
                 layer_sizes: list,
                 latent_dim: int,
                 use_bn: bool,
                 use_ln: bool,
                 use_dr: bool,
                 dr_rate: float,
                 num_classes: int = None):
        super().__init__()
        self.n_classes = 0
        if num_classes is not None:
            self.n_classes = num_classes
        self.FC = None
        if len(layer_sizes) > 1:
            print("Encoder Architecture:")
            self.FC = nn.Sequential()
            for i, (in_size, out_size) in enumerate(zip(layer_sizes[:-1], layer_sizes[1:])):
                if i == 0:
                    print("\tInput Layer in, out and cond:", in_size, out_size, self.n_classes)
                    self.FC.add_module(name="L{:d}".format(i), module=CondLayers(in_size,
                                                                                 out_size,
                                                                                 self.n_classes,
                                                                                 bias=True))
                else:
                    print("\tHidden Layer", i, "in/out:", in_size, out_size)
                    self.FC.add_module(name="L{:d}".format(i), module=nn.Linear(in_size, out_size, bias=True))
                if use_bn:
                    self.FC.add_module("N{:d}".format(i), module=nn.BatchNorm1d(out_size, affine=True))
                elif use_ln:
                    self.FC.add_module("N{:d}".format(i), module=nn.LayerNorm(out_size, elementwise_affine=False))
                self.FC.add_module(name="A{:d}".format(i), module=nn.ReLU())
                if use_dr:
                    self.FC.add_module(name="D{:d}".format(i), module=nn.Dropout(p=dr_rate))
        print("\tMean/Var Layer in/out:", layer_sizes[-1], latent_dim)
        self.mean_encoder = nn.Linear(layer_sizes[-1], latent_dim)
        self.log_var_encoder = nn.Linear(layer_sizes[-1], latent_dim)

    def forward(self, x, batch=None):
        if batch is not None:
            batch = one_hot_encoder(batch, n_cls=self.n_classes)
            x = torch.cat((x, batch), dim=-1)
        if self.FC is not None:
            x = self.FC(x)
        means = self.mean_encoder(x)
        log_vars = self.log_var_encoder(x)
        return means, log_vars


class Decoder(nn.Module):
    """ScArches Decoder class. Constructs the decoder sub-network of TRVAE or CVAE networks. It will transform the
       constructed latent space to the previous space of data with n_dimensions = x_dimension.

       Parameters
       ----------
       layer_sizes: List
            List of hidden and last layer sizes
       latent_dim: Integer
            Bottleneck layer (z)  size.
       recon_loss: String
            Definition of Reconstruction-Loss-Method, 'mse', 'nb' or 'zinb'.
       use_bn: Boolean
            If `True` batch normalization will be applied to layers.
       use_ln: Boolean
            If `True` layer normalization will be applied to layers.
       use_dr: Boolean
            If `True` dropout will applied to layers.
       dr_rate: Float
            Dropput rate applied to all layers, if `dr_rate`==0 no dropput will be applied.
       num_classes: Integer
            Number of classes (conditions) the data contain. if `None` the model will be a normal VAE instead of
            conditional VAE.
    """
    def __init__(self,
                 layer_sizes: list,
                 latent_dim: int,
                 recon_loss: str,
                 use_bn: bool,
                 use_ln: bool,
                 use_dr: bool,
                 dr_rate: float,
                 num_classes: int = None):
        super().__init__()
        self.use_dr = use_dr
        self.recon_loss = recon_loss
        self.n_classes = 0
        if num_classes is not None:
            self.n_classes = num_classes
        layer_sizes = [latent_dim] + layer_sizes
        print("Decoder Architecture:")
        # Create first Decoder layer
        self.FirstL = nn.Sequential()
        print("\tFirst Layer in, out and cond: ", layer_sizes[0], layer_sizes[1], self.n_classes)
        self.FirstL.add_module(name="L0", module=CondLayers(layer_sizes[0], layer_sizes[1], self.n_classes, bias=False))
        if use_bn:
            self.FirstL.add_module("N0", module=nn.BatchNorm1d(layer_sizes[1], affine=True))
        elif use_ln:
            self.FirstL.add_module("N0", module=nn.LayerNorm(layer_sizes[1], elementwise_affine=False))
        self.FirstL.add_module(name="A0", module=nn.ReLU())
        if self.use_dr:
            self.FirstL.add_module(name="D0", module=nn.Dropout(p=dr_rate))

        # Create all Decoder hidden layers
        if len(layer_sizes) > 2:
            self.HiddenL = nn.Sequential()
            for i, (in_size, out_size) in enumerate(zip(layer_sizes[1:-1], layer_sizes[2:])):
                if i+3 < len(layer_sizes):
                    print("\tHidden Layer", i+1, "in/out:", in_size, out_size)
                    self.HiddenL.add_module(name="L{:d}".format(i+1), module=nn.Linear(in_size, out_size, bias=False))
                    if use_bn:
                        self.HiddenL.add_module("N{:d}".format(i+1), module=nn.BatchNorm1d(out_size, affine=True))
                    elif use_ln:
                        self.HiddenL.add_module("N{:d}".format(i + 1), module=nn.LayerNorm(out_size, elementwise_affine=False))
                    self.HiddenL.add_module(name="A{:d}".format(i+1), module=nn.ReLU())
                    if self.use_dr:
                        self.HiddenL.add_module(name="D{:d}".format(i+1), module=nn.Dropout(p=dr_rate))
        else:
            self.HiddenL = None

        # Create Output Layers
        print("\tOutput Layer in/out: ", layer_sizes[-2], layer_sizes[-1], "\n")
        if self.recon_loss == "mse":
            self.recon_decoder = nn.Sequential(nn.Linear(layer_sizes[-2], layer_sizes[-1]), nn.ReLU())
        if self.recon_loss == "zinb":
            # mean gamma
            self.mean_decoder = nn.Sequential(nn.Linear(layer_sizes[-2], layer_sizes[-1]), nn.Softmax(dim=-1))
            # dropout
            self.dropout_decoder = nn.Linear(layer_sizes[-2], layer_sizes[-1])
        if self.recon_loss == "nb":
            # mean gamma
            self.mean_decoder = nn.Sequential(nn.Linear(layer_sizes[-2], layer_sizes[-1]), nn.Softmax(dim=-1))

    def forward(self, z, batch=None):
        # Add Condition Labels to Decoder Input
        if batch is not None:
            batch = one_hot_encoder(batch, n_cls=self.n_classes)
            z_cat = torch.cat((z, batch), dim=-1)
            dec_latent = self.FirstL(z_cat)
        else:
            dec_latent = self.FirstL(z)

        # Compute Hidden Output
        if self.HiddenL is not None:
            x = self.HiddenL(dec_latent)
        else:
            x = dec_latent

        # Compute Decoder Output
        if self.recon_loss == "mse":
            recon_x = self.recon_decoder(x)
            return recon_x, dec_latent
        elif self.recon_loss == "zinb":
            dec_mean_gamma = self.mean_decoder(x)
            dec_dropout = self.dropout_decoder(x)
            return dec_mean_gamma, dec_dropout, dec_latent
        elif self.recon_loss == "nb":
            dec_mean_gamma = self.mean_decoder(x)
            return dec_mean_gamma, dec_latent



import torch
from torch.utils.data import WeightedRandomSampler

from scarches.dataset.trvae import AnnotatedDataset

import re
from torch._six import container_abcs


import scanpy as sc

from scarches.trainers.trvae._utils import train_test_split

device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
print(device)

def custom_collate(batch):
    r"""Puts each data field into a tensor with outer dimension batch size"""
    np_str_obj_array_pattern = re.compile(r'[SaUO]')
    default_collate_err_msg_format = (
        "default_collate: batch must contain tensors, numpy arrays, numbers, "
        "dicts or lists; found {}")

    elem = batch[0]
    elem_type = type(elem)
    if isinstance(elem, torch.Tensor):
        out = None
        if torch.utils.data.get_worker_info() is not None:
            # If we're in a background process, concatenate directly into a
            # shared memory tensor to avoid an extra copy
            numel = sum([x.numel() for x in batch])
            storage = elem.storage()._new_shared(numel)
            out = elem.new(storage)
        return torch.stack(batch, 0, out=out)

    elif elem_type.__module__ == 'numpy' and elem_type.__name__ != 'str_' and elem_type.__name__ != 'string_':
        elem = batch[0]
        if elem_type.__name__ == 'ndarray' or elem_type.__name__ == 'memmap':
            # array of string classes and object
            if np_str_obj_array_pattern.search(elem.dtype.str) is not None:
                raise TypeError(default_collate_err_msg_format.format(elem.dtype))

            return custom_collate([torch.as_tensor(b) for b in batch])
        elif elem.shape == ():  # scalars
            return torch.as_tensor(batch)

    elif isinstance(elem, container_abcs.Mapping):
        if "y" in elem:
            output = dict(total=dict(),
                          labelled=dict())
            for key in elem:
                total_data = [d[key] for d in batch]
                labelled_data = list()
                for d in batch:
                    if d["y"] != -1:
                        labelled_data.append(d[key])
                output["total"][key] = custom_collate(total_data)
                output["labelled"][key] = custom_collate(labelled_data)
        else:
            output = dict(total=dict)
            output["total"] = {key: custom_collate([d[key] for d in batch]) for key in elem}
        return output

def make_dataset(adata,
                 train_frac=0.9,
                 condition_key="study",
                 cell_type_key="cell_type",
                 size_factor_key="size_factors",
                 is_label_key="is_labelled"):
    """Splits 'adata' into train and validation data and converts them into 'CustomDatasetFromAdata' objects.

       Parameters
       ----------

       Returns
       -------
       Training 'CustomDatasetFromAdata' object, Validation 'CustomDatasetFromAdata' object
    """
    train_adata, validation_adata = train_test_split(adata, train_frac)
    data_set_train = AnnotatedDataset(train_adata,
                                      condition_key=condition_key,
                                      cell_type_key=cell_type_key,
                                      size_factors_key=size_factor_key,
                                      )
    data_set_valid = AnnotatedDataset(validation_adata,
                                      condition_key=condition_key,
                                      cell_type_key=cell_type_key,
                                      size_factors_key=size_factor_key,
                                      )
    return data_set_train, data_set_valid

def inference(x=None, c=None, y=None, raw=None, f=None):
    for item in c.unique():
        print("Cell Item: ", item)
        iter = 0
        for i in range(c.size(0)):
            if c[i] == item:
                iter += 1
        print("occurences:", iter)

def forward(total=None, labelled=None):
    for key, batch in total.items():
        batch.to(device)
        print(batch)
    inference(**total)
    inference(**labelled)

def train_scannet(adata, n_epochs=5, batch_size=1024):
    dataset_train, dataset_valid = make_dataset(adata)
    strat_weights = dataset_train.stratifier_weights
    sampler = WeightedRandomSampler(strat_weights, 2048, replacement=True)
    data_loader_train = torch.utils.data.DataLoader(dataset=dataset_train,
                                                    batch_size=batch_size,
                                                    sampler=sampler,
                                                    collate_fn=custom_collate,
                                                    num_workers=0)
    for epoch in range(n_epochs):
        print("Epoch", epoch)
        for iteration, batch_data in enumerate(data_loader_train):
            print("Iteration: ", iteration)
            forward(**batch_data)

target_conditions = ["Pancreas SS2", "Pancreas CelSeq2"]
adata = sc.read("../../pancreas_normalized.h5ad", backup_url=f"https://zenodo.org/record/3930949/files/pancreas_normalized.h5ad?download=1", sparse=True, cache=True)

target_adata = adata[adata.obs["study"].isin(target_conditions)]
train_scannet(target_adata)



from .trainer import Trainer
import torch


class trVAETrainer(Trainer):
    """ScArches Unsupervised Trainer class. This class contains the implementation of the unsupervised CVAE/TRVAE
       Trainer.

           Parameters
           ----------
           model: trVAE
                Number of input features (i.e. gene in case of scRNA-seq).
           adata: : `~anndata.AnnData`
                Annotated data matrix.
           condition_key: String
                column name of conditions in `adata.obs` data frame.
           cell_type_key: String
                column name of celltypes in `adata.obs` data frame.
           train_frac: Float
                Defines the fraction of data that is used for training and data that is used for validation.
           batch_size: Integer
                Defines the batch size that is used during each Iteration
           n_samples: Integer or None
                Defines how many samples are being used during each epoch. This should only be used if hardware resources
                are limited.
           clip_value: Float
                If the value is greater than 0, all gradients with an higher value will be clipped during training.
           weight decay: Float
                Defines the scaling factor for weight decay in the Adam optimizer.
           alpha_iter_anneal: Integer or None
                If not 'None', the KL Loss scaling factor will be annealed from 0 to 1 every iteration until the input
                integer is reached.
           alpha_epoch_anneal: Integer or None
                If not 'None', the KL Loss scaling factor will be annealed from 0 to 1 every epoch until the input
                integer is reached.
           use_early_stopping: Boolean
                If 'True' the EarlyStopping class is being used for training to prevent overfitting.
           early_stopping_kwargs: Dict
                Passes custom Earlystopping parameters.
           use_stratified_sampling: Boolean
                If 'True', the sampler tries to load equally distributed batches concerning the conditions in every
                iteration.
           use_stratified_split: Boolean
                If `True`, the train and validation data will be constructed in such a way that both have same distribution
                of conditions in the data.
           monitor: Boolean
                If `True', the progress of the training will be printed after each epoch.
           n_workers: Integer
                Passes the 'n_workers' parameter for the torch.utils.data.DataLoader class.
           seed: Integer
                Define a specific random seed to get reproducable results.
        """
    def __init__(
            self,
            model,
            adata,
            **kwargs
    ):
        super().__init__(model, adata, **kwargs)

    def loss(self, total_batch=None):
        recon_loss, kl_loss, mmd_loss = self.model(**total_batch)
        loss = recon_loss + self.calc_alpha_coeff()*kl_loss + mmd_loss
        self.iter_logs["loss"].append(loss)
        self.iter_logs["unweighted_loss"].append(recon_loss + kl_loss + mmd_loss)
        self.iter_logs["recon_loss"].append(recon_loss)
        self.iter_logs["kl_loss"].append(kl_loss)
        if self.model.use_mmd:
            self.iter_logs["mmd_loss"].append(mmd_loss)
        return loss


from typing import Optional

import torch
import torch.nn as nn
from torch.distributions import Normal, kl_divergence
import torch.nn.functional as F

from .modules import Encoder, Decoder
from .losses import mse, mmd, zinb, nb
from ._utils import one_hot_encoder


class trVAE(nn.Module):
    """ScArches model class. This class contains the implementation of Conditional Variational Auto-encoder.

       Parameters
       ----------
       input_dim: Integer
            Number of input features (i.e. gene in case of scRNA-seq).
       conditions: List
            List of Condition names that the used data will contain to get the right encoding when used after reloading.
       hidden_layer_sizes: List
            A list of hidden layer sizes for encoder network. Decoder network will be the reversed order.
       latent_dim: Integer
            Bottleneck layer (z)  size.
       dr_rate: Float
            Dropput rate applied to all layers, if `dr_rate`==0 no dropout will be applied.
       use_mmd: Boolean
            If 'True' an additional MMD loss will be calculated on the latent dim. 'z' or the first decoder layer 'y'.
       mmd_on: String
            Choose on which layer MMD loss will be calculated on if 'use_mmd=True': 'z' for latent dim or 'y' for first
            decoder layer.
       mmd_boundary: Integer or None
            Choose on how many conditions the MMD loss should be calculated on. If 'None' MMD will be calculated on all
            conditions.
       recon_loss: String
            Definition of Reconstruction-Loss-Method, 'mse', 'nb' or 'zinb'.
       beta: Float
            Scaling Factor for MMD loss
       use_bn: Boolean
            If `True` batch normalization will be applied to layers.
       use_ln: Boolean
            If `True` layer normalization will be applied to layers.
    """

    def __init__(self,
                 input_dim: int,
                 conditions: list,
                 hidden_layer_sizes: list = [256, 64],
                 latent_dim: int = 10,
                 dr_rate: float = 0.05,
                 use_mmd: bool = False,
                 mmd_on: str = 'z',
                 mmd_boundary: Optional[int] = None,
                 recon_loss: Optional[str] = 'nb',
                 beta: float = 1,
                 use_bn: bool = False,
                 use_ln: bool = True,
                 ):
        super().__init__()
        assert isinstance(hidden_layer_sizes, list)
        assert isinstance(latent_dim, int)
        assert isinstance(conditions, list)
        assert recon_loss in ["mse", "nb", "zinb"], "'recon_loss' must be 'mse', 'nb' or 'zinb'"

        print("\nINITIALIZING NEW NETWORK..............")
        self.input_dim = input_dim
        self.latent_dim = latent_dim
        self.n_conditions = len(conditions)
        self.conditions = conditions
        self.condition_encoder = {k: v for k, v in zip(conditions, range(len(conditions)))}
        self.recon_loss = recon_loss
        self.mmd_boundary = mmd_boundary
        self.use_mmd = use_mmd
        self.freeze = False
        self.beta = beta
        self.use_bn = use_bn
        self.use_ln = use_ln
        self.mmd_on = mmd_on

        self.dr_rate = dr_rate
        if self.dr_rate > 0:
            self.use_dr = True
        else:
            self.use_dr = False

        if recon_loss in ["nb", "zinb"]:
            self.theta = torch.nn.Parameter(torch.randn(self.input_dim, self.n_conditions))
        else:
            self.theta = None

        self.hidden_layer_sizes = hidden_layer_sizes
        encoder_layer_sizes = self.hidden_layer_sizes.copy()
        encoder_layer_sizes.insert(0, self.input_dim)
        decoder_layer_sizes = self.hidden_layer_sizes.copy()
        decoder_layer_sizes.reverse()
        decoder_layer_sizes.append(self.input_dim)
        self.encoder = Encoder(encoder_layer_sizes,
                               self.latent_dim,
                               self.use_bn,
                               self.use_ln,
                               self.use_dr,
                               self.dr_rate,
                               self.n_conditions)
        self.decoder = Decoder(decoder_layer_sizes,
                               self.latent_dim,
                               self.recon_loss,
                               self.use_bn,
                               self.use_ln,
                               self.use_dr,
                               self.dr_rate,
                               self.n_conditions)

    def sampling(self, mu, log_var):
        """Samples from standard Normal distribution and applies re-parametrization trick.
           It is actually sampling from latent space distributions with N(mu, var), computed by encoder.

           Parameters
           ----------
           mu: torch.Tensor
                Torch Tensor of Means.
           log_var: torch.Tensor
                Torch Tensor of log. variances.

           Returns
           -------
           Torch Tensor of sampled data.
        """
        var = torch.exp(log_var) + 1e-4
        return Normal(mu, var.sqrt()).rsample()

    def get_latent(self, x, c=None, mean=False):
        """Map `x` in to the latent space. This function will feed data in encoder  and return  z for each sample in
           data.

           Parameters
           ----------
           x:  torch.Tensor
                Torch Tensor to be mapped to latent space. `x` has to be in shape [n_obs, input_dim].
           c: torch.Tensor
                Torch Tensor of condition labels for each sample.
           mean: boolean

           Returns
           -------
                Returns array containing latent space encoding of 'x'.
        """
        x_ = torch.log(1 + x)
        z_mean, z_log_var = self.encoder(x_, c)
        latent = self.sampling(z_mean, z_log_var)
        if mean:
            return z_mean
        return latent

    def get_y(self, x, c=None):
        """Map `x` in to the y dimension (First Layer of Decoder). This function will feed data in encoder  and return
           y for each sample in data.

           Parameters
           ----------
           x:  torch.Tensor
                Torch Tensor to be mapped to latent space. `x` has to be in shape [n_obs, input_dim].
           c: torch.Tensor
                Torch Tensor of condition labels for each sample.

           Returns
           -------
           Returns array containing latent space encoding of 'x'
        """
        z_mean, z_log_var = self.encoder(x, c)
        latent = self.sampling(z_mean, z_log_var)
        output = self.decoder(latent, c)
        return output[-1]

    def forward(self, x=None, batch=None, sizefactor=None):
        x_log = torch.log(1 + x)
        z1_mean, z1_log_var = self.encoder(x_log, batch)
        z1 = self.sampling(z1_mean, z1_log_var)
        outputs = self.decoder(z1, batch)

        if self.recon_loss == "mse":
            recon_x, y1 = outputs
            recon_loss = mse(recon_x, x_log).sum(dim=-1).mean()
        elif self.recon_loss == "zinb":
            dec_mean_gamma, dec_dropout, y1 = outputs
            size_factor_view = sizefactor.unsqueeze(1).expand(dec_mean_gamma.size(0), dec_mean_gamma.size(1))
            dec_mean = dec_mean_gamma * size_factor_view
            dispersion = F.linear(one_hot_encoder(batch, self.n_conditions), self.theta)
            dispersion = torch.exp(dispersion)
            recon_loss = -zinb(x=x, mu=dec_mean, theta=dispersion, pi=dec_dropout).sum(dim=-1).mean()
        elif self.recon_loss == "nb":
            dec_mean_gamma, y1 = outputs
            size_factor_view = sizefactor.unsqueeze(1).expand(dec_mean_gamma.size(0), dec_mean_gamma.size(1))
            dec_mean = dec_mean_gamma * size_factor_view
            dispersion = F.linear(one_hot_encoder(batch, self.n_conditions), self.theta)
            dispersion = torch.exp(dispersion)
            recon_loss = -nb(x=x, mu=dec_mean, theta=dispersion).sum(dim=-1).mean()

        z1_var = torch.exp(z1_log_var) + 1e-4
        kl_div = kl_divergence(
            Normal(z1_mean, torch.sqrt(z1_var)),
            Normal(torch.zeros_like(z1_mean), torch.ones_like(z1_var))
        ).sum(dim=1).mean()

        mmd_loss = 0
        if self.use_mmd:
            mmd_calculator = mmd(self.n_conditions, self.beta, self.mmd_boundary)
            if self.mmd_on == "z":
                mmd_loss = mmd_calculator(z1, batch)
            else:
                mmd_loss = mmd_calculator(y1, batch)

        return recon_loss, kl_div, mmd_loss



from scvi.data import get_from_registry

from scarches.metrics.metrics import entropy_batch_mixing, knn_purity, asw, nmi
from scarches.models import SCVI, SCANVI, TOTALVI
from scarches.trainers import scVITrainer, scANVITrainer, totalTrainer

from scipy.sparse import issparse
import numpy as np
import scanpy as sc
import torch
from typing import Union
from sklearn.metrics import f1_score
import anndata
import matplotlib.pyplot as plt

sc.settings.set_figure_params(dpi=200, frameon=False)
torch.set_printoptions(precision=3, sci_mode=False, edgeitems=7)
np.set_printoptions(precision=2, edgeitems=7)


class SCVI_EVAL:
    def __init__(
            self,
            model: Union[SCVI, SCANVI, TOTALVI],
            adata: anndata.AnnData,
            trainer: Union[scVITrainer, scANVITrainer, totalTrainer] = None,
            cell_type_key: str = None,
            batch_key: str = None,
    ):
        self.outer_model = model
        self.model = model.model
        self.model.eval()

        if trainer is None:
            self.trainer = model.trainer
        else:
            self.trainer = trainer

        self.adata = adata
        self.modified = getattr(model.model, 'encode_covariates', True)
        self.annotated = type(model) is SCANVI
        self.predictions = None
        self.certainty = None
        self.prediction_names = None
        self.class_check = None
        self.post_adata_2 = None

        if trainer is not None:
            if self.trainer.use_cuda:
                self.device = torch.device('cuda')
            else:
                self.device = torch.device('cpu')
        else:
            self.device = next(self.model.parameters()).get_device()

        if issparse(self.adata.X):
            X = self.adata.X.toarray()
        else:
            X = self.adata.X
        self.x_tensor = torch.tensor(X, device=self.device)

        self.labels = None
        self.label_tensor = None
        if self.annotated:
            self.labels = get_from_registry(self.adata, "labels").astype(np.int8)
            self.label_tensor = torch.tensor(self.labels, device=self.device)
        self.cell_types = self.adata.obs[cell_type_key].tolist()

        self.batch_indices = get_from_registry(self.adata, "batch_indices").astype(np.int8)
        self.batch_tensor = torch.tensor(self.batch_indices, device=self.device)
        self.batch_names = self.adata.obs[batch_key].tolist()
        self.celltype_enc = [0]*len(self.adata.obs[cell_type_key].unique().tolist())
        for i, cell_type in enumerate(self.adata.obs[cell_type_key].unique().tolist()):
            label = self.adata.obs['_scvi_labels'].unique().tolist()[i]
            self.celltype_enc[label] = cell_type
        self.post_adata = self.latent_as_anndata()

    def latent_as_anndata(self):
        if type(self.outer_model) is TOTALVI:
            latent = self.outer_model.get_latent_representation(self.adata)
        else:
            if self.modified:
                latents = self.model.sample_from_posterior_z(
                    self.x_tensor,
                    y=self.label_tensor,
                    batch_index=self.batch_tensor
                )
            else:
                latents = self.model.sample_from_posterior_z(
                    self.x_tensor,
                    y=self.label_tensor,
                )

            if self.annotated:
                latent = latents.cpu().detach().numpy()
                latent2, _, _ = self.model.encoder_z2_z1(latents, self.label_tensor)
                latent2 = latent2.cpu().detach().numpy()
                post_adata_2 = sc.AnnData(latent2)
                post_adata_2.obs['cell_type'] = self.cell_types
                post_adata_2.obs['batch'] = self.batch_names
                self.post_adata_2 = post_adata_2
            else:
                latent = latents.cpu().detach().numpy()

        post_adata = sc.AnnData(latent)
        post_adata.obs['cell_type'] = self.cell_types
        post_adata.obs['batch'] = self.batch_names
        return post_adata

    def get_model_arch(self):
        for name, p in self.model.named_parameters():
            print(name, " - ", p.size(0), p.size(-1))

    def plot_latent(self,
                    show=True,
                    save=False,
                    dir_path=None,
                    n_neighbors=8,
                    predictions=False,
                    in_one=False,
                    colors=None):
        """
        if save:
            if dir_path is None:
                name = 'scanvi_latent.png'
            else:
                name = f'{dir_path}.png'
        else:
            name = False
        """
        if self.model is None:
            print("Not possible if no model is provided")
            return
        if save:
            show = False
            if dir_path is None:
                dir_path = 'scanvi_latent'
        sc.pp.neighbors(self.post_adata, n_neighbors=n_neighbors)
        sc.tl.umap(self.post_adata)
        if in_one:
            color = ['cell_type', 'batch']
            if predictions:
                color.append(['certainty', 'predictions', 'type_check'])
            sc.pl.umap(self.post_adata,
                       color=color,
                       ncols=2,
                       frameon=False,
                       wspace=0.6,
                       show=show,
                       save=f'{dir_path}_complete.png' if dir_path else None)
        else:
            sc.pl.umap(self.post_adata,
                       color=['cell_type'],
                       frameon=False,
                       wspace=0.6,
                       show=show,
                       palette=colors,
                       save=f'{dir_path}_celltypes.png' if dir_path else None)
            sc.pl.umap(self.post_adata,
                       color=['batch'],
                       frameon=False,
                       wspace=0.6,
                       show=show,
                       save=f'{dir_path}_batch.png' if dir_path else None)
            if predictions:
                sc.pl.umap(self.post_adata,
                           color=['predictions'],
                           frameon=False,
                           wspace=0.6,
                           show=show,
                           save=f'{dir_path}_predictions.png' if dir_path else None)
                sc.pl.umap(self.post_adata,
                           color=['certainty'],
                           frameon=False,
                           wspace=0.6,
                           show=show,
                           save=f'{dir_path}_certainty.png' if dir_path else None)
                sc.pl.umap(self.post_adata,
                           color=['type_check'],
                           ncols=2,
                           frameon=False,
                           wspace=0.6,
                           show=show,
                           save=f'{dir_path}_type_check.png' if dir_path else None)

    def plot_history(self, show=True, save=False, dir_path=None):
        if self.trainer is None:
            print("Not possible if no trainer is provided")
            return
        if self.annotated:
            fig, axs = plt.subplots(2, 1)
            elbo_full = self.trainer.history["elbo_full_dataset"]
            x_1 = np.linspace(0, len(elbo_full), len(elbo_full))
            axs[0].plot(x_1, elbo_full, label="Full")

            accuracy_labelled_set = self.trainer.history["accuracy_labelled_set"]
            accuracy_unlabelled_set = self.trainer.history["accuracy_unlabelled_set"]
            if len(accuracy_labelled_set) != 0:
                x_2 = np.linspace(0, len(accuracy_labelled_set), (len(accuracy_labelled_set)))
                axs[1].plot(x_2, accuracy_labelled_set, label="accuracy labelled")
            if len(accuracy_unlabelled_set) != 0:
                x_3 = np.linspace(0, len(accuracy_unlabelled_set), (len(accuracy_unlabelled_set)))
                axs[1].plot(x_3, accuracy_unlabelled_set, label="accuracy unlabelled")
            axs[0].set_xlabel('Epochs')
            axs[0].set_ylabel('ELBO')
            axs[1].set_xlabel('Epochs')
            axs[1].set_ylabel('Accuracy')
            axs[0].legend(bbox_to_anchor=(1.05, 1), loc='upper left', borderaxespad=0.)
            axs[1].legend(bbox_to_anchor=(1.05, 1), loc='upper left', borderaxespad=0.)
            if save:
                if dir_path is None:
                    plt.savefig('scanvi_history.png', bbox_inches='tight')
                else:
                    plt.savefig(f'{dir_path}.png', bbox_inches='tight')
            if show:
                plt.show()
        else:
            fig = plt.figure()
            elbo_train = self.trainer.history["elbo_train_set"]
            elbo_test = self.trainer.history["elbo_test_set"]
            x = np.linspace(0, len(elbo_train), len(elbo_train))
            plt.plot(x, elbo_train, label="train")
            plt.plot(x, elbo_test, label="test")
            plt.ylim(min(elbo_train) - 50, min(elbo_train) + 1000)
            plt.legend(bbox_to_anchor=(1.05, 1), loc='upper left', borderaxespad=0.)
            if save:
                if dir_path is None:
                    plt.savefig('scvi_history.png', bbox_inches='tight')
                else:
                    plt.savefig(f'{dir_path}.png', bbox_inches='tight')
            if show:
                plt.show()

    def get_ebm(self, n_neighbors=50, n_pools=50, n_samples_per_pool=100, verbose=True):
        ebm_score = entropy_batch_mixing(
            adata=self.post_adata,
            label_key='batch',
            n_neighbors=n_neighbors,
            n_pools=n_pools,
            n_samples_per_pool=n_samples_per_pool
        )
        if verbose:
            print("Entropy of Batchmixing-Score:", ebm_score)
        return ebm_score

    def get_knn_purity(self, n_neighbors=50, verbose=True):
        knn_score = knn_purity(
            adata=self.post_adata,
            label_key='cell_type',
            n_neighbors=n_neighbors
        )
        if verbose:
            print("KNN Purity-Score:", knn_score)
        return knn_score

    def get_asw(self):
        asw_score_batch, asw_score_cell_types = asw(adata=self.post_adata, label_key='cell_type',batch_key='batch')
        print("ASW on batch:", asw_score_batch)
        print("ASW on celltypes:", asw_score_cell_types)
        return asw_score_batch, asw_score_cell_types

    def get_nmi(self):
        nmi_score = nmi(adata=self.post_adata, label_key='cell_type')
        print("NMI score:", nmi_score)
        return nmi_score

    def get_latent_score(self):
        ebm = self.get_ebm(verbose=False)
        knn = self.get_knn_purity(verbose=False)
        score = ebm + knn
        print("Latent-Space Score (KNN + EBM):", score)
        return score

    def get_classification_accuracy(self):
        if self.annotated:
            if self.modified:
                softmax = self.model.classify(self.x_tensor, batch_index=self.batch_tensor)
            else:
                softmax = self.model.classify(self.x_tensor)
            softmax = softmax.cpu().detach().numpy()
            self.predictions = np.argmax(softmax, axis=1)
            self.certainty = np.max(softmax, axis=1)
            self.prediction_names = [0]*self.predictions.shape[0]
            for index, label in np.ndenumerate(self.predictions):
                self.prediction_names[index[0]] = self.celltype_enc[label]
            self.class_check = np.array(np.expand_dims(self.predictions, axis=1) == self.labels)
            class_check_labels = [0] * self.class_check.shape[0]
            for index, check in np.ndenumerate(self.class_check):
                class_check_labels[index[0]] = 'Correct' if check else 'Incorrect'
            accuracy = np.sum(self.class_check) / self.class_check.shape[0]
            self.post_adata.obs['certainty'] = self.certainty
            self.post_adata.obs['type_check'] = class_check_labels
            self.post_adata.obs['predictions'] = self.prediction_names
            print("Classification Accuracy: %0.2f" % accuracy)
            return accuracy
        else:
            print("Classification ratio not available for scVI models")

    def get_f1_score(self):
        if self.annotated:
            if self.modified:
                predictions = self.model.classify(self.x_tensor, batch_index=self.batch_tensor)
            else:
                predictions = self.model.classify(self.x_tensor)
            self.predictions = predictions.cpu().detach().numpy()
            self.predictions = np.expand_dims(np.argmax(self.predictions, axis=1), axis=1)
            score = f1_score(self.labels, self.predictions, average='macro')
            print("F1 Score: %0.2f" % score)
            return score
        else:
            print("F1 Score not available for scVI models")



import numpy as np


class EarlyStopping(object):
    """Class for EarlyStopping. This class contains the implementation of early stopping for TRVAE/CVAE training.

           Parameters
           ----------
           early_stopping_metric: : String
                The metric/loss which the early stopping criterion gets calculated on.
           threshold: Float
                The minimum value which counts as improvement.
           patience: Integer
                Number of epochs which are allowed to have no improvement until the training is stopped.
           reduce_lr: Boolean
                If 'True', the learning rate gets adjusted by 'lr_factor' after a given number of epochs with no
                improvement.
           lr_patience: Integer
                Number of epochs which are allowed to have no improvement until the learning rate is adjusted.
           lr_factor: Float
                Scaling factor for adjusting the learning rate.
        """
    def __init__(self,
                 early_stopping_metric: str = 'val_unweighted_loss',
                 threshold: float = 0,
                 patience: int = 20,
                 reduce_lr: bool = True,
                 lr_patience: int = 7,
                 lr_factor: float = 0.1):

        self.early_stopping_metric = early_stopping_metric
        self.threshold = threshold
        self.patience = patience
        self.reduce_lr = reduce_lr
        self.lr_patience = lr_patience
        self.lr_factor = lr_factor

        self.epoch = 0
        self.wait = 0
        self.wait_lr = 0
        self.current_performance = np.inf
        self.best_performance = np.inf
        self.best_performance_state = np.inf

        if patience == 0:
            self.is_better = lambda a, b: True
            self.step = lambda a: False

    def step(self, scalar):
        self.epoch += 1
        if self.epoch < self.patience:
            continue_training = True
            lr_update = False
        elif self.wait >= self.patience:
            continue_training = False
            lr_update = False
        else:
            if not self.reduce_lr:
                lr_update = False
            elif self.wait_lr >= self.lr_patience:
                lr_update = True
                self.wait_lr = 0
            else:
                lr_update = False
            # Shift
            self.current_performance = scalar

            improvement = self.best_performance - self.current_performance

            # updating best performance
            if improvement > 0:
                self.best_performance = self.current_performance

            if improvement < self.threshold:
                self.wait += 1
                self.wait_lr += 1
            else:
                self.wait = 0
                self.wait_lr = 0

            continue_training = True

        if not continue_training:
            print("\nStopping early: no improvement of more than " + str(self.threshold) +
                  " nats in " + str(self.patience) + " epochs")
            print("If the early stopping criterion is too strong, "
                  "please instantiate it with different parameters in the train method.")
        return continue_training, lr_update

    def update_state(self, scalar):
        improved = (self.best_performance_state - scalar) > 0
        if improved:
            self.best_performance_state = scalar
        return improved



import numpy as np
import pandas as pd
from scipy.stats import itemfreq, entropy
from sklearn.metrics import silhouette_score, normalized_mutual_info_score, silhouette_samples
from sklearn.neighbors import NearestNeighbors
from sklearn.preprocessing import LabelEncoder

from scarches.dataset.trvae.data_handling import remove_sparsity
from .clustering import opt_louvain


def entropy_batch_mixing(adata, label_key='batch',
                         n_neighbors=50, n_pools=50, n_samples_per_pool=100):
    """Computes Entory of Batch mixing metric for ``adata`` given the batch column name.
        Parameters
        ----------
        adata: :class:`~anndata.AnnData`
            Annotated dataset.
        label_key: str
            Name of the column which contains information about different studies in ``adata.obs`` data frame.
        n_neighbors: int
            Number of nearest neighbors.
        n_pools: int
            Number of EBM computation which will be averaged.
        n_samples_per_pool: int
            Number of samples to be used in each pool of execution.
        Returns
        -------
        score: float
            EBM score. A float between zero and one.
    """
    adata = remove_sparsity(adata)
    n_cat = len(adata.obs[label_key].unique().tolist())
    print(f'Calculating EBM with n_cat = {n_cat}')

    neighbors = NearestNeighbors(n_neighbors=n_neighbors + 1).fit(adata.X)
    indices = neighbors.kneighbors(adata.X, return_distance=False)[:, 1:]
    batch_indices = np.vectorize(lambda i: adata.obs[label_key].values[i])(indices)

    entropies = np.apply_along_axis(__entropy_from_indices, axis=1, arr=batch_indices, n_cat=n_cat)

    # average n_pools entropy results where each result is an average of n_samples_per_pool random samples.
    if n_pools == 1:
        score = np.mean(entropies)
    else:
        score = np.mean([
            np.mean(entropies[np.random.choice(len(entropies), size=n_samples_per_pool)])
            for _ in range(n_pools)
        ])

    return score


def nmi(adata, label_key, verbose=False, nmi_method='arithmetic'):
    cluster_key = 'cluster'
    opt_louvain(adata, label_key=label_key, cluster_key=cluster_key, function=nmi_helper,
                plot=False, verbose=verbose, inplace=True)

    print('NMI...')
    nmi_score = nmi_helper(adata, group1=cluster_key, group2=label_key, method=nmi_method)

    return nmi_score


def asw(adata, label_key, batch_key):
    print('silhouette score...')
    sil_global = silhouette(adata, group_key=label_key, metric='euclidean')
    _, sil_clus = silhouette_batch(adata, batch_key=batch_key, group_key=label_key,
                                   metric='euclidean', verbose=False)
    sil_clus = sil_clus['silhouette_score'].mean()
    return sil_clus, sil_global


def knn_purity(adata, label_key, n_neighbors=30):
    """Computes KNN Purity metric for ``adata`` given the batch column name.
        Parameters
        ----------
        adata: :class:`~anndata.AnnData`
            Annotated dataset.
        label_key: str
            Name of the column which contains information about different studies in ``adata.obs`` data frame.
        n_neighbors: int
            Number of nearest neighbors.
        Returns
        -------
        score: float
            KNN purity score. A float between 0 and 1.
    """
    adata = remove_sparsity(adata)
    labels = LabelEncoder().fit_transform(adata.obs[label_key].to_numpy())

    nbrs = NearestNeighbors(n_neighbors=n_neighbors + 1).fit(adata.X)
    indices = nbrs.kneighbors(adata.X, return_distance=False)[:, 1:]
    neighbors_labels = np.vectorize(lambda i: labels[i])(indices)

    # pre cell purity scores
    scores = ((neighbors_labels - labels.reshape(-1, 1)) == 0).mean(axis=1)
    res = [
        np.mean(scores[labels == i]) for i in np.unique(labels)
    ]  # per cell-type purity

    return np.mean(res)


def __entropy_from_indices(indices, n_cat):
    return entropy(np.array(itemfreq(indices)[:, 1].astype(np.int32)), base=n_cat)


def nmi_helper(adata, group1, group2, method="arithmetic"):
    """
    Normalized mutual information NMI based on 2 different cluster assignments `group1` and `group2`
    params:
        adata: Anndata object
        group1: column name of `adata.obs` or group assignment
        group2: column name of `adata.obs` or group assignment
        method: NMI implementation
            'max': scikit method with `average_method='max'`
            'min': scikit method with `average_method='min'`
            'geometric': scikit method with `average_method='geometric'`
            'arithmetic': scikit method with `average_method='arithmetic'`
            'Lancichinetti': implementation by A. Lancichinetti 2009 et al.
            'ONMI': implementation by Aaron F. McDaid et al. (https://github.com/aaronmcdaid/Overlapping-NMI) Hurley 2011
        nmi_dir: directory of compiled C code if 'Lancichinetti' or 'ONMI' are specified as `method`. Compilation should be done as specified in the corresponding README.
    return:
        normalized mutual information (NMI)
    """
    adata = remove_sparsity(adata)

    if isinstance(group1, str):
        group1 = adata.obs[group1].tolist()
    elif isinstance(group1, pd.Series):
        group1 = group1.tolist()

    labels = adata.obs[group2].values
    labels_encoded = LabelEncoder().fit_transform(labels)
    group2 = labels_encoded

    if len(group1) != len(group2):
        raise ValueError(f'different lengths in group1 ({len(group1)}) and group2 ({len(group2)})')

    # choose method
    if method in ['max', 'min', 'geometric', 'arithmetic']:
        nmi_value = normalized_mutual_info_score(group1, group2, average_method=method)
    else:
        raise ValueError(f"Method {method} not valid")

    return nmi_value


def silhouette(adata, group_key, metric='euclidean', scale=True):
    """
    wrapper for sklearn silhouette function values range from [-1, 1] with 1 being an ideal fit, 0 indicating overlapping clusters and -1 indicating misclassified cells
    """
    adata = remove_sparsity(adata)
    labels = adata.obs[group_key].values
    labels_encoded = LabelEncoder().fit_transform(labels)
    asw = silhouette_score(adata.X, labels_encoded, metric=metric)
    if scale:
        asw = (asw + 1)/2
    return asw


def silhouette_batch(adata, batch_key, group_key, metric='euclidean', verbose=True, scale=True):
    """
    Silhouette score of batch labels subsetted for each group.
    params:
        batch_key: batches to be compared against
        group_key: group labels to be subsetted by e.g. cell type
        metric: see sklearn silhouette score
        embed: name of column in adata.obsm
    returns:
        all scores: absolute silhouette scores per group label
        group means: if `mean=True`
    """
    adata = remove_sparsity(adata)
    glob_batches = adata.obs[batch_key].values
    batch_enc = LabelEncoder()
    batch_enc.fit(glob_batches)
    sil_all = pd.DataFrame(columns=['group', 'silhouette_score'])

    for group in adata.obs[group_key].unique():
        adata_group = adata[adata.obs[group_key] == group]
        if adata_group.obs[batch_key].nunique() == 1:
            continue
        batches = batch_enc.transform(adata_group.obs[batch_key])
        sil_per_group = silhouette_samples(adata_group.X, batches, metric=metric)
        # take only absolute value
        sil_per_group = [abs(i) for i in sil_per_group]
        if scale:
            # scale s.t. highest number is optimal
            sil_per_group = [1 - i for i in sil_per_group]
        d = pd.DataFrame({'group': [group] * len(sil_per_group), 'silhouette_score': sil_per_group})
        sil_all = sil_all.append(d)
    sil_all = sil_all.reset_index(drop=True)
    sil_means = sil_all.groupby('group').mean()

    if verbose:
        print(f'mean silhouette per cell: {sil_means}')
    return sil_all, sil_means



import torch
import torch.nn as nn
from collections import defaultdict
import numpy as np
import time
from torch.utils.data import DataLoader
from torch.utils.data import WeightedRandomSampler

from scarches.utils.monitor import EarlyStopping
from ._utils import make_dataset, custom_collate, print_progress


class Trainer:
    """ScArches base Trainer class. This class contains the implementation of the base CVAE/TRVAE Trainer.

       Parameters
       ----------
       model: trVAE
            Number of input features (i.e. gene in case of scRNA-seq).
       adata: : `~anndata.AnnData`
            Annotated data matrix.
       condition_key: String
            column name of conditions in `adata.obs` data frame.
       cell_type_key: String
            column name of celltypes in `adata.obs` data frame.
       train_frac: Float
            Defines the fraction of data that is used for training and data that is used for validation.
       batch_size: Integer
            Defines the batch size that is used during each Iteration
       n_samples: Integer or None
            Defines how many samples are being used during each epoch. This should only be used if hardware resources
            are limited.
       clip_value: Float
            If the value is greater than 0, all gradients with an higher value will be clipped during training.
       weight decay: Float
            Defines the scaling factor for weight decay in the Adam optimizer.
       alpha_iter_anneal: Integer or None
            If not 'None', the KL Loss scaling factor will be annealed from 0 to 1 every iteration until the input
            integer is reached.
       alpha_epoch_anneal: Integer or None
            If not 'None', the KL Loss scaling factor will be annealed from 0 to 1 every epoch until the input
            integer is reached.
       use_early_stopping: Boolean
            If 'True' the EarlyStopping class is being used for training to prevent overfitting.
       early_stopping_kwargs: Dict
            Passes custom Earlystopping parameters.
       use_stratified_sampling: Boolean
            If 'True', the sampler tries to load equally distributed batches concerning the conditions in every
            iteration.
       use_stratified_split: Boolean
            If `True`, the train and validation data will be constructed in such a way that both have same distribution
            of conditions in the data.
       monitor: Boolean
            If `True', the progress of the training will be printed after each epoch.
       n_workers: Integer
            Passes the 'n_workers' parameter for the torch.utils.data.DataLoader class.
       seed: Integer
            Define a specific random seed to get reproducable results.
    """
    def __init__(self,
                 model,
                 adata,
                 condition_key: str = None,
                 cell_type_key: str = None,
                 train_frac: float = 0.9,
                 batch_size: int = 128,
                 n_samples: int = None,
                 clip_value: float = 0.0,
                 weight_decay: float = 0.04,
                 alpha_iter_anneal: int = None,
                 alpha_epoch_anneal: int = None,
                 early_stopping_kwargs: dict = None,
                 **kwargs):

        self.adata = adata
        self.model = model
        self.condition_key = condition_key
        self.cell_type_key = cell_type_key
        self.train_frac = train_frac

        self.batch_size = batch_size
        self.n_samples = n_samples
        self.clip_value = clip_value
        self.weight_decay = weight_decay
        self.alpha_iter_anneal = alpha_iter_anneal
        self.alpha_epoch_anneal = alpha_epoch_anneal

        early_stopping_kwargs = (early_stopping_kwargs if early_stopping_kwargs else dict())

        self.use_early_stopping = kwargs.pop("use_early_stopping", True)
        self.use_stratified_sampling = kwargs.pop("use_stratified_sampling", True)
        self.use_stratified_split = kwargs.pop("use_stratified_split", False)
        self.monitor = kwargs.pop("monitor", True)
        self.n_workers = kwargs.pop("n_workers", 0)
        self.seed = kwargs.pop("seed", 2020)

        self.early_stopping = EarlyStopping(**early_stopping_kwargs)
        torch.manual_seed(self.seed)
        if torch.cuda.is_available():
            torch.cuda.manual_seed(self.seed)
            self.model.cuda()
        self.device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')

        self.epoch = -1
        self.n_epochs = None
        self.iter = 0
        self.best_epoch = None
        self.best_state_dict = None
        self.current_loss = None
        self.previous_loss_was_nan = False
        self.nan_counter = 0
        self.optimizer = None
        self.training_time = 0

        self.train_data = None
        self.valid_data = None
        self.sampler = None
        self.dataloader_train = None
        self.dataloader_valid = None

        self.iters_per_epoch = None
        self.val_iters_per_epoch = None

        self.logs = defaultdict(list)

    def calc_alpha_coeff(self):
        """Calculates current alpha coefficient for alpha annealing.

           Parameters
           ----------

           Returns
           -------
           Current annealed alpha value
        """
        if self.alpha_epoch_anneal is not None:
            alpha_coeff = min(self.epoch / self.alpha_epoch_anneal, 1)
        elif self.alpha_iter_anneal is not None:
            alpha_coeff = min(((self.epoch * self.iters_per_epoch + self.iter) / self.alpha_iter_anneal), 1)
        else:
            alpha_coeff = 1
        return alpha_coeff

    def train(self,
              n_epochs=400,
              lr=1e-3,
              eps=0.01):
        begin = time.time()
        self.model.train()
        self.n_epochs = n_epochs

        params = filter(lambda p: p.requires_grad, self.model.parameters())

        self.optimizer = torch.optim.Adam(params, lr=lr, eps=eps, weight_decay=self.weight_decay)
        # Initialize Train/Val Data, Sampler, Dataloader
        self.initialize_loaders()

        for self.epoch in range(n_epochs):
            self.iter_logs = defaultdict(list)
            for self.iter, batch_data in enumerate(self.dataloader_train):
                # Safe data to right device
                for key1 in batch_data:
                    for key2, batch in batch_data[key1].items():
                        batch_data[key1][key2] = batch.to(self.device)

                # Loss Calculation
                self.on_iteration(batch_data)

            # Validation of Model, Monitoring, Early Stopping
            self.on_epoch_end()
            if self.use_early_stopping:
                if not self.check_early_stop():
                    break

        if self.best_state_dict is not None:
            print("Saving best state of network...")
            print("Best State was in Epoch", self.best_epoch)
            self.model.load_state_dict(self.best_state_dict)

        self.model.eval()

        self.training_time += (time.time() - begin)

    def initialize_loaders(self):
        """
        Initializes Train-/Test Data and Dataloaders with custom_collate and WeightedRandomSampler for Trainloader.
        Returns:

        """
        use_normalized = False
        if self.model.recon_loss == 'mse':
            use_normalized = True

        # Create Train/Valid AnnotatetDataset objects
        self.train_data, self.valid_data = make_dataset(
            self.adata,
            train_frac=self.train_frac,
            use_stratified_split=self.use_stratified_split,
            condition_key=self.condition_key,
            cell_type_key=self.cell_type_key,
            condition_encoder=self.model.condition_encoder,
            cell_type_encoder=None,
            use_normalized=use_normalized,
        )

        if self.n_samples is None or self.n_samples > len(self.train_data):
            self.n_samples = len(self.train_data)
        self.iters_per_epoch = int(np.ceil(self.n_samples / self.batch_size))

        if self.use_stratified_sampling:
            # Create Sampler and Dataloaders
            stratifier_weights = torch.tensor(self.train_data.stratifier_weights, device=self.device)

            self.sampler = WeightedRandomSampler(stratifier_weights,
                                                 num_samples=self.n_samples,
                                                 replacement=True)
            self.dataloader_train = torch.utils.data.DataLoader(dataset=self.train_data,
                                                                batch_size=self.batch_size,
                                                                sampler=self.sampler,
                                                                collate_fn=custom_collate,
                                                                num_workers=self.n_workers)
        else:
            self.dataloader_train = torch.utils.data.DataLoader(dataset=self.train_data,
                                                                batch_size=self.batch_size,
                                                                shuffle=True,
                                                                collate_fn=custom_collate,
                                                                num_workers=self.n_workers)
        if self.valid_data is not None:
            print("Valid_data", len(self.valid_data))
            for i in range(len(self.valid_data.unique_conditions)):
                print("Condition:", i, "Counts in TrainData:", np.count_nonzero(self.valid_data.conditions == i))

            val_batch_size = self.batch_size
            if self.batch_size > len(self.valid_data):
                val_batch_size = len(self.valid_data)
            self.val_iters_per_epoch = int(np.ceil(len(self.valid_data) / self.batch_size))
            self.dataloader_valid = torch.utils.data.DataLoader(dataset=self.valid_data,
                                                                batch_size=val_batch_size,
                                                                shuffle=True,
                                                                collate_fn=custom_collate,
                                                                num_workers=self.n_workers)

    def on_iteration(self, batch_data):
        # Dont update any weight on first layers except condition weights
        if self.model.freeze:
            for name, module in self.model.named_modules():
                if isinstance(module, nn.BatchNorm1d):
                    if not module.weight.requires_grad:
                        module.affine = False
                        module.track_running_stats = False

        # Calculate Loss depending on Trainer/Model
        self.current_loss = loss = self.loss(**batch_data)
        self.optimizer.zero_grad()
        loss.backward()

        # Gradient Clipping
        if self.clip_value > 0:
            torch.nn.utils.clip_grad_value_(self.model.parameters(), self.clip_value)

        self.optimizer.step()

    def on_epoch_end(self):
        # Get Train Epoch Logs
        for key in self.iter_logs:
            if "loss" in key:
                self.logs["epoch_" + key].append(
                    sum(self.iter_logs[key][:]).cpu().detach().numpy() / len(self.iter_logs[key][:]))

        # Validate Model
        if self.valid_data is not None:
            self.validate()

        # Monitor Logs
        if self.monitor:
            print_progress(self.epoch, self.logs, self.n_epochs)

    @torch.no_grad()
    def validate(self):
        self.model.eval()
        self.iter_logs = defaultdict(list)
        # Calculate Validation Losses
        for val_iter, batch_data in enumerate(self.dataloader_valid):
            for key1 in batch_data:
                for key2, batch in batch_data[key1].items():
                    batch_data[key1][key2] = batch.to(self.device)

            val_loss = self.loss(**batch_data)

        # Get Validation Logs
        for key in self.iter_logs:
            if "loss" in key:
                self.logs["val_" + key].append(
                    sum(self.iter_logs[key][:]).cpu().detach().numpy() / len(self.iter_logs[key][:]))

        self.model.train()

    def check_early_stop(self):
        # Calculate Early Stopping and best state
        early_stopping_metric = self.early_stopping.early_stopping_metric
        if self.early_stopping.update_state(self.logs[early_stopping_metric][-1]):
            self.best_state_dict = self.model.state_dict()
            self.best_epoch = self.epoch

        continue_training, update_lr = self.early_stopping.step(self.logs[early_stopping_metric][-1])
        if update_lr:
            print(f'\nADJUSTED LR')
            for param_group in self.optimizer.param_groups:
                param_group["lr"] *= self.early_stopping.lr_factor

        return continue_training



import sys
import numpy as np
import re
import torch
from torch._six import container_abcs
from torch.utils.data import DataLoader, SubsetRandomSampler

from scarches.dataset.trvae import AnnotatedDataset


def print_progress(epoch, logs, n_epochs=10000):
    """Creates Message for '_print_progress_bar'.

       Parameters
       ----------
       epoch: Integer
            Current epoch iteration.
       logs: dict
            Dictionary of all current losses.
       n_epochs: Integer
            Maximum value of epochs.

       Returns
       -------
    """
    message = ""
    for key in logs:
        if "loss" in key and ("epoch_" in key or "val_" in key):
            message += f" - {key:s}: {logs[key][-1]:7.0f}"

    _print_progress_bar(epoch + 1, n_epochs, prefix='', suffix=message, decimals=1, length=20)


def _print_progress_bar(iteration, total, prefix='', suffix='', decimals=1, length=100, fill='█'):
    """Prints out message with a progress bar.

       Parameters
       ----------
       iteration: Integer
            Current epoch.
       total: Integer
            Maximum value of epochs.
       prefix: String
            String before the progress bar.
       suffix: String
            String after the progress bar.
       decimals: Integer
            Digits after comma for all the losses.
       length: Integer
            Length of the progress bar.
       fill: String
            Symbol for filling the bar.

       Returns
       -------
    """
    percent = ("{0:." + str(decimals) + "f}").format(100 * (iteration / float(total)))
    filled_len = int(length * iteration // total)
    bar = fill * filled_len + '-' * (length - filled_len)
    sys.stdout.write('\r%s |%s| %s%s %s' % (prefix, bar, percent, '%', suffix)),
    if iteration == total:
        sys.stdout.write('\n')
    sys.stdout.flush()


def train_test_split(adata, train_frac=0.85, condition_key=None):
    """Splits 'Anndata' object into training and validation data.

       Parameters
       ----------
       adata: `~anndata.AnnData`
            `AnnData` object for training the model.
       train_frac: float
            Train-test split fraction. the model will be trained with train_frac for training
            and 1-train_frac for validation.

       Returns
       -------
       `AnnData` objects for training and validating the model.
    """
    if train_frac == 1:
        return adata, None
    else:
        indices = np.arange(adata.shape[0])
        n_val_samples = int(adata.shape[0] * (1 - train_frac))
        if condition_key is not None:
            conditions = adata.obs[condition_key].unique().tolist()
            n_conditions = len(conditions)
            n_val_samples_per_condition = int(n_val_samples / n_conditions)
            condition_indices_train = []
            condition_indices_val = []
            for i in range(n_conditions):
                idx = indices[adata.obs[condition_key] == conditions[i]]
                np.random.shuffle(idx)
                condition_indices_val.append(idx[:n_val_samples_per_condition])
                condition_indices_train.append(idx[n_val_samples_per_condition:])
            train_idx = np.concatenate(condition_indices_train)
            val_idx = np.concatenate(condition_indices_val)
        else:
            np.random.shuffle(indices)
            val_idx = indices[:n_val_samples]
            train_idx = indices[n_val_samples:]

        train_data = adata[train_idx, :]
        valid_data = adata[val_idx, :]

        return train_data, valid_data


def make_dataset(adata,
                 train_frac=0.9,
                 use_stratified_split=False,
                 condition_key=None,
                 cell_type_key=None,
                 condition_encoder=None,
                 cell_type_encoder=None,
                 use_normalized=False
                 ):
    """Splits 'adata' into train and validation data and converts them into 'CustomDatasetFromAdata' objects.

       Parameters
       ----------

       Returns
       -------
       Training 'CustomDatasetFromAdata' object, Validation 'CustomDatasetFromAdata' object
    """
    if use_stratified_split:
        train_adata, validation_adata = train_test_split(adata, train_frac, condition_key=condition_key)
    else:
        train_adata, validation_adata = train_test_split(adata, train_frac)

    data_set_train = AnnotatedDataset(train_adata,
                                      condition_key=condition_key,
                                      cell_type_key=cell_type_key,
                                      condition_encoder=condition_encoder,
                                      cell_type_encoder=cell_type_encoder,
                                      use_normalized=use_normalized,
                                      )
    if train_frac == 1:
        return data_set_train, None
    else:
        data_set_valid = AnnotatedDataset(validation_adata,
                                          condition_key=condition_key,
                                          cell_type_key=cell_type_key,
                                          condition_encoder=condition_encoder,
                                          cell_type_encoder=cell_type_encoder,
                                          use_normalized=use_normalized,
                                          )
        return data_set_train, data_set_valid


def custom_collate(batch):
    r"""Puts each data field into a tensor with outer dimension batch size"""
    np_str_obj_array_pattern = re.compile(r'[SaUO]')
    default_collate_err_msg_format = (
        "default_collate: batch must contain tensors, numpy arrays, numbers, "
        "dicts or lists; found {}")

    elem = batch[0]
    elem_type = type(elem)
    if isinstance(elem, torch.Tensor):
        out = None
        if torch.utils.data.get_worker_info() is not None:
            # If we're in a background process, concatenate directly into a
            # shared memory tensor to avoid an extra copy
            numel = sum([x.numel() for x in batch])
            storage = elem.storage()._new_shared(numel)
            out = elem.new(storage)
        return torch.stack(batch, 0, out=out)

    elif elem_type.__module__ == 'numpy' and elem_type.__name__ != 'str_' and elem_type.__name__ != 'string_':
        elem = batch[0]
        if elem_type.__name__ == 'ndarray' or elem_type.__name__ == 'memmap':
            # array of string classes and object
            if np_str_obj_array_pattern.search(elem.dtype.str) is not None:
                raise TypeError(default_collate_err_msg_format.format(elem.dtype))

            return custom_collate([torch.as_tensor(b) for b in batch])
        elif elem.shape == ():  # scalars
            return torch.as_tensor(batch)

    elif isinstance(elem, container_abcs.Mapping):
        if "celltype" in elem:
            output = dict(total_batch=dict(),
                          labelled_batch=dict())
            for key in elem:
                total_data = [d[key] for d in batch]
                labelled_data = list()
                for d in batch:
                    if d["celltype"] != -1:
                        labelled_data.append(d[key])
                output["total_batch"][key] = custom_collate(total_data)
                output["labelled_batch"][key] = custom_collate(labelled_data)
        else:
            output = dict(total_batch=dict())
            output["total_batch"] = {key: custom_collate([d[key] for d in batch]) for key in elem}
        return output



import numpy as np
import scanpy as sc
import torch
import anndata
import matplotlib.pyplot as plt
from typing import Union

from scarches.dataset.trvae._utils import label_encoder
from scarches.metrics.metrics import entropy_batch_mixing, knn_purity, asw, nmi
from scarches.models import trVAE, TRVAE
from scarches.trainers import trVAETrainer

sc.settings.set_figure_params(dpi=200, frameon=False)
sc.set_figure_params(dpi=200)
torch.set_printoptions(precision=3, sci_mode=False, edgeitems=7)
np.set_printoptions(precision=2, edgeitems=7)


class TRVAE_EVAL:
    def __init__(
            self,
            model: Union[trVAE, TRVAE],
            adata: anndata.AnnData,
            trainer: trVAETrainer = None,
            condition_key: str = None,
            cell_type_key: str = None
    ):
        if type(model) is TRVAE:
            trainer = model.trainer
            model = model.model

        self.model = model
        self.trainer = trainer
        self.adata = adata
        self.device = model.device
        self.conditions, _ = label_encoder(
            self.adata,
            encoder=model.condition_encoder,
            condition_key=condition_key,
        )
        self.cell_type_names = None
        self.batch_names = None
        if cell_type_key is not None:
            self.cell_type_names = adata.obs[cell_type_key].tolist()
        if condition_key is not None:
            self.batch_names = adata.obs[condition_key].tolist()

        self.adata_latent = self.latent_as_anndata()

    def latent_as_anndata(self):
        if self.model.calculate_mmd == 'z' or self.model.use_mmd == False:
            latent = self.model.get_latent(
                self.adata.X,
                c=self.conditions,
            )
        else:
            latent = self.model.get_y(
                self.adata.X,
                c=self.conditions
            )
        adata_latent = sc.AnnData(latent)
        if self.cell_type_names is not None:
            adata_latent.obs['cell_type'] = self.cell_type_names
        if self.batch_names is not None:
            adata_latent.obs['batch'] = self.batch_names
        return adata_latent

    def get_model_arch(self):
        for name, p in self.model.named_parameters():
            print(name, " - ", p.size(0), p.size(-1))

    def plot_latent(self,
                    show=True,
                    save=False,
                    dir_path=None,
                    n_neighbors=8,
                    ):
        if save:
            show=False
            if dir_path is None:
                save = False

        sc.pp.neighbors(self.adata_latent, n_neighbors=n_neighbors)
        sc.tl.umap(self.adata_latent)
        color = [
            'cell_type' if self.cell_type_names is not None else None,
            'batch' if self.batch_names is not None else None,
        ]
        sc.pl.umap(self.adata_latent,
                   color=color,
                   frameon=False,
                   wspace=0.6,
                   show=show)
        if save:
            plt.savefig(f'{dir_path}_batch.png', bbox_inches='tight')

    def plot_history(self, show=True, save=False, dir_path=None):
        if save:
            show = False
            if dir_path is None:
                save = False

        if self.trainer is None:
            print("Not possible if no trainer is provided")
            return
        fig = plt.figure()
        elbo_train = self.trainer.logs["epoch_loss"]
        elbo_test = self.trainer.logs["val_loss"]
        x = np.linspace(0, len(elbo_train), num=len(elbo_train))
        plt.plot(x, elbo_train, label="Train")
        plt.plot(x, elbo_test, label="Validate")
        plt.ylim(min(elbo_test) - 50, max(elbo_test) + 50)
        plt.legend()
        if save:
            plt.savefig(f'{dir_path}.png', bbox_inches='tight')
        if show:
            plt.show()
        plt.clf()

    def get_ebm(self, n_neighbors=50, n_pools=50, n_samples_per_pool=100, verbose=True):
        ebm_score = entropy_batch_mixing(
            adata=self.adata_latent,
            label_key='batch',
            n_neighbors=n_neighbors,
            n_pools=n_pools,
            n_samples_per_pool=n_samples_per_pool
        )
        if verbose:
            print("Entropy of Batchmixing-Score: %0.2f" % ebm_score)
        return ebm_score

    def get_knn_purity(self, n_neighbors=50, verbose=True):
        knn_score = knn_purity(
            adata=self.adata_latent,
            label_key='cell_type',
            n_neighbors=n_neighbors
        )
        if verbose:
            print("KNN Purity-Score:  %0.2f" % knn_score)
        return knn_score

    def get_asw(self):
        asw_score_batch, asw_score_cell_types = asw(adata=self.adata_latent, label_key='cell_type', batch_key='batch')
        print("ASW on batch:", asw_score_batch)
        print("ASW on celltypes:", asw_score_cell_types)
        return asw_score_batch, asw_score_cell_types

    def get_nmi(self):
        nmi_score = nmi(adata=self.adata_latent, label_key='cell_type')
        print("NMI score:", nmi_score)
        return nmi_score

    def get_latent_score(self):
        ebm = self.get_ebm(verbose=False)
        knn = self.get_knn_purity(verbose=False)
        score = ebm + knn
        print("Latent-Space Score EBM+KNN, EBM, KNN: %0.2f, %0.2f, %0.2f" % (score, ebm, knn))
        return score



from . import metrics, models, dataset, trainers, zenodo, plotting
from .surgery import scvi_operate, trvae_operate

__author__ = ', '.join([
    'Marco Wagenstetter',
    'Mohsen Naghipourfar',
    'Mohammad Lotfollahi',
    'Sergei Rybakov'
])

__email__ = ', '.join([
    'marco.wagenstetter@tum.de',
    'naghipourfar@ce.sharif.edu',
    'mohammad.lotfollahi@helmholtz-muenchen.de',
    'sergei.rybakov@helmholtz-muenchen.de'
])



from typing import Union

from ..models import TRVAE, SCVI, SCANVI, TOTALVI
from .file import *
from .deposition import *
from .zip import *


def upload_model(model: Union[TRVAE, SCVI, SCANVI, TOTALVI, str],
                 deposition_id: str,
                 access_token: str,
                 model_name: str = None):
    """Uploads trained ``model`` to Zenodo.

        Parameters
        ----------
        model: :class:`~scarches.models.TRVAE`, :class:`~scarches.models.SCVI`, :class:`~scarches.models.SCANVI`, :class:`~scarches.models.TOTALVI`, str
            An instance of one of classes defined in ``scarches.models`` module or a path to a saved model.
        deposition_id: str
            ID of a deposition in your Zenodo account.
        access_token: str
            Your Zenodo access token.
        model_name: str
            An optional name of the model to upload

        Returns
        -------
        download_link: str
            Generated direct download link for the uploaded model in the deposition. Please **Note** that the link is usable **after** your published your deposition.
    """
    if model_name is None:
        model_name = type(model).__name__

    if isinstance(model, str):
        model_path = model
    else:
        model_path = f"tmp_{model_name}"
        model.save(model_path)

    output_base_name = f"./tmp/scarches-{model_name}"
    output_path = output_base_name + ".zip"
    zip_model_directory(output_path=output_base_name, directory=model_path)
    download_link = upload_file(file_path=output_path,
                                deposition_id=deposition_id,
                                access_token=access_token)
    print("Model has been successfully uploaded")
    return download_link


def download_model(download_link: str,
                   save_path: str = './',
                   make_dir: bool = False):
    """Downloads the zip file of the model in the ``link`` and saves it in ``save_path`` and extracts.

        Parameters
        ----------
        link: str
            Direct downloadable link.
        save_path: str
            Directory path for downloaded file
        make_dir: bool
            Whether to make the ``save_path`` if it does not exist in the system.

        Returns
        -------
        extract_dir: str
            Full path to the folder of the model.
    """
    if not save_path.endswith("/"):
        save_path += "/"
    if download_link != '':
        file_path, response = download_file(download_link, f'{save_path}downloaded_model.zip', make_dir)
    else:
        raise Exception("Download link does not exist for the specified task")

    if os.path.exists(file_path) and file_path.endswith(".zip"):
        extract_dir = os.path.dirname(file_path)
        unzip_model_directory(file_path, extract_dir=extract_dir)
    else:
        raise Exception("The model should be in zip archive")

    return extract_dir



import numpy as np
import torch
from torch.utils.data import Dataset
from scipy import sparse
import scanpy as sc

from .data_handling import remove_sparsity
from ._utils import label_encoder


class AnnotatedDataset(Dataset):
    """Dataset handler for TRVAE model and trainer.
       Parameters
       ----------
       adata: : `~anndata.AnnData`
            Annotated data matrix.
       condition_key: String
            column name of conditions in `adata.obs` data frame.
       condition_encoder: Dict or None
            dictionary of encoded conditions. if `None`, will create one.
       cell_type_key: String
            column name of celltypes in `adata.obs` data frame.
       cell_type_encoder: Dict or None
            dictionary of encoded celltypes. if `None`, will create one.
       unique_conditions: List or None
            List of all conditions in the data.
       use_normalized: Boolean
            If 'True', adata.X is getting normalized and used as input for the network, this is only necessary for
            MSE loss.
    """
    def __init__(self,
                 adata,
                 condition_key=None,
                 condition_encoder=None,
                 cell_type_key=None,
                 cell_type_encoder=None,
                 unique_conditions=None,
                 use_normalized=False,
                 ):

        # Desparse Adata
        self.adata = adata
        if sparse.issparse(self.adata.X):
            self.adata = remove_sparsity(self.adata)

        self.use_normalized = use_normalized
        self.X_norm = None

        self.condition_key = condition_key
        self.condition_encoder = condition_encoder
        self.unique_conditions = unique_conditions

        self.cell_type_key = cell_type_key
        self.cell_type_encoder = cell_type_encoder
        self.unique_cell_types = None

        if self.use_normalized:
            sc.pp.normalize_total(adata,
                                  exclude_highly_expressed=True,
                                  target_sum=1e4,
                                  key_added='size_factors')
        else:
            size_factors = np.log(adata.X.sum(1))
            if len(size_factors.shape) < 2:
                size_factors = np.expand_dims(size_factors, axis=1)
            adata.obs['size_factors'] = size_factors

        # Create Condition Encoder
        if self.condition_key is not None:
            if self.unique_conditions is None:
                self.unique_conditions = adata.obs[condition_key].unique().tolist()
            self.conditions, self.condition_encoder = label_encoder(self.adata,
                                                                    encoder=self.condition_encoder,
                                                                    condition_key=condition_key)
            self.conditions = np.array(self.conditions).reshape(-1, )

        # Create Label Encoder
        if self.cell_type_key is not None:
            if self.unique_cell_types is None:
                self.unique_cell_types = adata.obs[cell_type_key].unique().tolist()
            self.cell_types, self.cell_type_encoder = label_encoder(self.adata,
                                                                    encoder=self.cell_type_encoder,
                                                                    condition_key=cell_type_key)
            self.cell_types = np.array(self.cell_types).reshape(-1, )

    def __getitem__(self, index):
        outputs = dict()

        outputs["x"] = torch.tensor(self.adata.X[index, :])

        if not self.use_normalized:
            outputs["sizefactor"] = torch.tensor(self.adata.obs['size_factors'][index])

        if self.condition_key:
            outputs["batch"] = torch.tensor(self.conditions[index])

        if self.cell_type_key:
            outputs["celltype"] = torch.tensor(self.cell_types[index])

        return outputs

    def __len__(self):
        return len(self.adata)

    @property
    def condition_label_encoder(self) -> dict:
        return self.condition_encoder

    @condition_label_encoder.setter
    def condition_label_encoder(self, value: dict):
        if value is not None:
            self.condition_encoder = value

    @property
    def cell_type_label_encoder(self) -> dict:
        return self.cell_type_encoder

    @cell_type_label_encoder.setter
    def cell_type_label_encoder(self, value: dict):
        if value is not None:
            self.cell_type_encoder = value

    @property
    def stratifier_weights(self):
        condition_coeff = 1 / len(self.conditions)
        weights_per_condition = list()
        for i in range(len(self.conditions)):
            samples_per_condition = np.count_nonzero(self.conditions == i)
            if samples_per_condition == 0:
                weights_per_condition.append(0)
            else:
                weights_per_condition.append((1 / samples_per_condition) * condition_coeff)
        strat_weights = np.copy(self.conditions)
        for i in range(len(self.conditions)):
            strat_weights = np.where(strat_weights == i, weights_per_condition[i], strat_weights)

        return strat_weights.astype(float)



import scanpy as sc
import pandas as pd
import seaborn as sns
import matplotlib.pyplot as plt
from scarches.dataset.trvae.data_handling import remove_sparsity


def opt_louvain(adata, label_key, cluster_key, function=None, resolutions=None,
                inplace=True, plot=False, verbose=True, **kwargs):
    """
    params:
        label_key: name of column in adata.obs containing biological labels to be
            optimised against
        cluster_key: name of column to be added to adata.obs during clustering.
            Will be overwritten if exists and `force=True`
        function: function that computes the cost to be optimised over. Must take as
            arguments (adata, group1, group2, **kwargs) and returns a number for maximising
        resolutions: list if resolutions to be optimised over. If `resolutions=None`,
            default resolutions of 20 values ranging between 0.1 and 2 will be used
    returns:
        res_max: resolution of maximum score
        score_max: maximum score
        score_all: `pd.DataFrame` containing all scores at resolutions. Can be used to plot the score profile.
        clustering: only if `inplace=False`, return cluster assignment as `pd.Series`
        plot: if `plot=True` plot the score profile over resolution
    """
    adata = remove_sparsity(adata)

    if resolutions is None:
        n = 20
        resolutions = [2 * x / n for x in range(1, n + 1)]

    score_max = 0
    res_max = resolutions[0]
    clustering = None
    score_all = []

    # maren's edit - recompute neighbors if not existing
    try:
        adata.uns['neighbors']
    except KeyError:
        if verbose:
            print('computing neigbours for opt_cluster')
        sc.pp.neighbors(adata)

    for res in resolutions:
        sc.tl.louvain(adata, resolution=res, key_added=cluster_key)
        score = function(adata, label_key, cluster_key, **kwargs)
        score_all.append(score)
        if score_max < score:
            score_max = score
            res_max = res
            clustering = adata.obs[cluster_key]
        del adata.obs[cluster_key]

    if verbose:
        print(f'optimised clustering against {label_key}')
        print(f'optimal cluster resolution: {res_max}')
        print(f'optimal score: {score_max}')

    score_all = pd.DataFrame(zip(resolutions, score_all), columns=('resolution', 'score'))
    if plot:
        # score vs. resolution profile
        sns.lineplot(data=score_all, x='resolution', y='score').set_title('Optimal cluster resolution profile')
        plt.show()

    if inplace:
        adata.obs[cluster_key] = clustering
        return res_max, score_max, score_all
    else:
        return res_max, score_max, score_all, clustering


import numpy as np
from typing import Union, List, Dict
import torch
import anndata

from scarches.models import SCVI, SCANVI, TOTALVI
from scarches.trainers import scVITrainer, scANVITrainer, totalTrainer

from scvi.data import transfer_anndata_setup
from scvi.core.models._utils import _initialize_model
from scvi.core.models.archesmixin import _set_params_online_update

def _transfer_model(model, adata):
    adata = adata.copy()

    attr_dict = model._get_user_attributes()
    attr_dict = {a[0]: a[1] for a in attr_dict if a[0][-1] == "_"}
    scvi_setup_dict = attr_dict.pop("scvi_setup_dict_")
    transfer_anndata_setup(scvi_setup_dict, adata, extend_categories=True)

    adata.uns["_scvi"]["summary_stats"]["n_labels"] = scvi_setup_dict[
        "summary_stats"
    ]["n_labels"]

    new_model = _initialize_model(model.__class__, adata, attr_dict, use_cuda=True)

    for attr, val in attr_dict.items():
        setattr(new_model, attr, val)

    model.model.cuda()
    new_model.model.cuda()
    new_state_dict = model.model.state_dict()

    load_state_dict = model.model.state_dict().copy()
    new_state_dict = new_model.model.state_dict()
    for key, load_ten in load_state_dict.items():
        new_ten = new_state_dict[key]
        if new_ten.size() == load_ten.size():
            continue
        else:
            dim_diff = new_ten.size()[-1] - load_ten.size()[-1]
            fixed_ten = torch.cat([load_ten, new_ten[..., -dim_diff:]], dim=-1)
            load_state_dict[key] = fixed_ten
    new_model.model.load_state_dict(load_state_dict)
    new_model.model.eval()

    new_model.is_trained_ = False

    return new_model, adata

def scvi_operate(
        model: Union[SCVI, SCANVI, TOTALVI],
        data: anndata.AnnData,
        labels_per_class: int = 0,
        n_epochs: int = 50,
        learning_rate: float = 1e-3,
        early_stopping: bool = False,
        freeze: bool = True,
        freeze_expression: bool = True,
        freeze_batchnorm_encoder: bool = True,
        freeze_batchnorm_decoder: bool = False,
        freeze_dropout: bool = False,
        **kwargs
) -> [Union[SCVI, SCANVI, TOTALVI], Union[scVITrainer, scANVITrainer, totalTrainer], anndata.AnnData]:
    """Transfer Learning function for new data. Uses old trained Network and expands it for new conditions.
       Parameters
       ----------
       model: VAE_M, SCANVI_M, TOTALVI_M
            A Scvi/Scanvi/Totalvi model object
       data: AnnData
            AnnData object.
       labels_per_class: Integer
            Number of labelled Samples used for Retraining
       n_epochs: Integer
            Number of epochs for training the network on query data.
       freeze: Boolean
            If 'True' freezes every part of the network except predefined layers dependent of model version.
       freeze_expression: Boolean
            If 'True' freeze every weight except the condition weights.
       freeze_batchnorm_encoder: Boolean
            If 'True' freeze Batchnorm in the encoder for Transfer Learning.
       freeze_batchnorm_decoder: Boolean
            If 'True' freeze Batchnorm in the decoder for Transfer Learning.
       freeze_dropout: Boolean
            If 'True' remove Dropout for Transfer Learning.

       Returns
       -------
       new_model: SCVI, SCANVI, TOTALVI
            Newly network that got trained on query data.
       op_trainer: scVITrainer, scANVITrainer, totalTrainer
            Trainer for the newly network.
       data: AnnData
            AnnData object with updated batch labels.
    """

    if not early_stopping:
        early_stopping_kwargs = None
    else:
        if type(model) is not SCANVI:
            early_stopping_kwargs = {
                "early_stopping_metric": "elbo",
                "save_best_state_metric": "elbo",
                "patience": 15,
                "threshold": 0,
                "reduce_lr_on_plateau": True,
                "lr_patience": 8,
                "lr_factor": 0.1,
            }
        else:
            early_stopping_kwargs = {
                "early_stopping_metric": "accuracy",
                "save_best_state_metric": "accuracy",
                "on": "full_dataset",
                "patience": 15,
                "threshold": 0.001,
                "reduce_lr_on_plateau": True,
                "lr_patience": 8,
                "lr_factor": 0.1,
            }

    new_model, adata = _transfer_model(model, data)
    if freeze:
        _set_params_online_update(
            new_model.model,
            freeze_batchnorm_encoder=freeze_batchnorm_encoder,
            freeze_batchnorm_decoder=freeze_batchnorm_decoder,
            freeze_dropout=freeze_dropout,
            freeze_expression=freeze_expression,
        )

    # Retrain Networks
    if type(new_model) is SCVI:
        op_trainer = scVITrainer(
            new_model,
            train_size=0.9,
            use_cuda=True,
            frequency=1,
            early_stopping_kwargs=early_stopping_kwargs,
            **kwargs
        )
    if type(new_model) is SCANVI:
        op_trainer = scANVITrainer(
            new_model,
            n_labelled_samples_per_class=labels_per_class,
            train_size=0.9,
            use_cuda=True,
            frequency=1,
            early_stopping_kwargs=early_stopping_kwargs,
            **kwargs
        )
    if type(new_model) is TOTALVI:
        op_trainer = totalTrainer(
            new_model,
            train_size=0.9,
            use_cuda=True,
            frequency=1,
            early_stopping_kwargs=early_stopping_kwargs,
            **kwargs
        )

    op_trainer.train(n_epochs=n_epochs, lr=learning_rate)

    return new_model, op_trainer, adata



import scanpy as sc
import torch
import os
import scarches as sca
from scarches.plotting import TRVAE_EVAL

save_path = os.path.expanduser(f'~/Documents/benchmark_results/trvae_umaps/testing/')
if not os.path.exists(save_path):
    os.makedirs(save_path)

results = dict()

sc.settings.set_figure_params(dpi=200, frameon=False)
sc.set_figure_params(dpi=200)
torch.set_printoptions(precision=3, sci_mode=False, edgeitems=7)

condition_key = 'study'
size_factor_key = 'size_factors'
cell_type_key = 'cell_type'
show_plots = False
trvae_epochs = 200
surgery_epochs = 100

target_conditions = ["Pancreas SS2", "Pancreas CelSeq2"]
adata = sc.read(os.path.expanduser(f'~/Documents/datasets/pancreas_normalized.h5ad'))
source_adata = adata[~adata.obs[condition_key].isin(target_conditions)]
target_adata = adata[adata.obs[condition_key].isin(target_conditions)]
source_conditions = source_adata.obs[condition_key].unique().tolist()

network = sca.models.trVAE(
    source_adata.n_vars,
    conditions=source_conditions,
    hidden_layer_sizes=[256, 128, 32],
    latent_dim=10,
    dr_rate=0.05,
    use_mmd=True,
    recon_loss="zinb",
)

trainer = sca.trainers.trVAETrainer(
    network,
    source_adata,
    alpha_epoch_anneal=200,
    condition_key=condition_key,
    size_factor_key=size_factor_key,
    batch_size=1024,
    n_samples=4096
)

trainer.train(
    n_epochs=trvae_epochs,
    lr=0.001
)

trvae_eval = TRVAE_EVAL(
    model=network,
    adata=source_adata,
    trainer=trainer,
    condition_key=condition_key,
    cell_type_key=cell_type_key
)
results["reference_ebm"] = trvae_eval.get_ebm()
results["reference_knn"] = trvae_eval.get_knn_purity()
results["reference_asw_b"], results["reference_asw_c"] = trvae_eval.get_asw()
results["reference_nmi"] = trvae_eval.get_nmi()
trvae_eval.plot_history(show=show_plots, save=True, dir_path=f'{save_path}reference_history')
trvae_eval.plot_latent(show=show_plots, save=True, dir_path=f'{save_path}reference_latent')
trvae_eval.adata_latent.write_h5ad(filename=f'{save_path}reference_latent_data.h5ad')
torch.save(network.state_dict(), f'{save_path}reference_model')

# --------------------------------------------------TRANSFER LEARNING---------------------------------------------------
new_network, new_trainer = sca.trvae_operate(
    network,
    target_adata,
    condition_key=condition_key,
    size_factor_key=size_factor_key,
    n_epochs=surgery_epochs,
    freeze=True,
    freeze_expression=True,
    remove_dropout=False
)
surgery_eval = TRVAE_EVAL(
    model=new_network,
    adata=target_adata,
    trainer=new_trainer,
    condition_key=condition_key,
    cell_type_key=cell_type_key
)
results["query_ebm"] = surgery_eval.get_ebm()
results["query_knn"] = surgery_eval.get_knn_purity()
results["query_asw_b"], results["query_asw_c"] = surgery_eval.get_asw()
results["query_nmi"] = surgery_eval.get_nmi()
surgery_eval.plot_history(show=show_plots, save=True, dir_path=f'{save_path}surgery_history')
surgery_eval.plot_latent(show=show_plots, save=True, dir_path=f'{save_path}query_latent')
surgery_eval.adata_latent.write_h5ad(filename=f'{save_path}query_latent_data.h5ad')
torch.save(new_network.state_dict(), f'{save_path}surgery_model')

# -------------------------------------------------Plotting Whole Data--------------------------------------------------
full_eval = TRVAE_EVAL(
    model=new_network,
    adata=adata,
    condition_key=condition_key,
    cell_type_key=cell_type_key
)

full_eval.get_latent_score()
if show_plots:
    full_eval.plot_latent()
full_eval = TRVAE_EVAL(
    model=new_network,
    adata=adata,
    trainer=new_trainer,
    condition_key=condition_key,
    cell_type_key=cell_type_key
)
results["full_ebm"] = full_eval.get_ebm()
results["full_knn"] = full_eval.get_knn_purity()
results["full_asw_b"], results["full_asw_c"] = full_eval.get_asw()
results["full_nmi"] = full_eval.get_nmi()
full_eval.plot_latent(show=show_plots, save=True, dir_path=f'{save_path}full_latent')
full_eval.adata_latent.write_h5ad(filename=f'{save_path}full_latent_data.h5ad')



from typing import Union
from scvi.core.trainers import UnsupervisedTrainer, SemiSupervisedTrainer, TotalTrainer


class scVITrainer(UnsupervisedTrainer):
    def __init__(
        self,
        scvi_model,
        train_size: Union[int, float] = 0.9,
        test_size: Union[int, float] = None,
        *args,
        **kwargs
    ):
        model = scvi_model.model
        adata = scvi_model.adata

        train_size = float(train_size)
        if train_size > 1.0 or train_size <= 0.0:
            raise ValueError(
                "train_size needs to be greater than 0 and less than or equal to 1"
            )

        super().__init__(model, adata, train_size, test_size, *args, **kwargs)

        self.train_set, self.test_set, self.validation_set = self.train_test_validation(
            model, adata, train_size, test_size
        )
        self.train_set.to_monitor = ["elbo"]
        self.test_set.to_monitor = ["elbo"]
        self.validation_set.to_monitor = ["elbo"]
        self.n_samples = len(self.train_set.indices)

        self._scvi_model = scvi_model

    def train(self, *args, **kwargs):
        super().train(*args, **kwargs)
        self._scvi_model.trainer = self
        self._scvi_model.is_trained_ = True


class scANVITrainer(SemiSupervisedTrainer):
    def __init__(self, scanvi_model, n_labelled_samples_per_class=50, *args, **kwargs):
        model = scanvi_model.model
        adata = scanvi_model.adata

        super().__init__(model, adata, n_labelled_samples_per_class, *args, **kwargs)

        self._scanvi_model = scanvi_model
        self._labelled = n_labelled_samples_per_class > 0

    def loss(self, tensors_all, tensors_labelled=None):
        if self._labelled:
            return SemiSupervisedTrainer.loss(self, tensors_all, tensors_labelled)
        else:
            return UnsupervisedTrainer.loss(self, tensors_all)

    def train(self, *args, **kwargs):
        super().train(*args, **kwargs)

        self._scanvi_model.trainer = self
        self._scanvi_model.is_trained_ = True

    @property
    def scvi_data_loaders_loop(self):
        if self._labelled:
            return ["full_dataset", "labelled_set"]
        else:
            return ["full_dataset"]


class totalTrainer(TotalTrainer):
    def __init__(self, totalvi_model, *args, **kwargs):
        model = totalvi_model.model
        adata = totalvi_model.adata

        # adapive use_adversarial_loss
        if len(args) < 7:
            if 'use_adversarial_loss' not in kwargs:
                if 'totalvi_batch_mask' in totalvi_model.scvi_setup_dict_:
                    imputation = True
                else:
                    imputation = False
                kwargs['use_adversarial_loss'] = imputation

        super().__init__(model, adata, *args, **kwargs)

        self._totalvi_model = totalvi_model

    def train(self, *args, **kwargs):
        super().train(*args, **kwargs)

        self._totalvi_model.trainer = self
        self._totalvi_model.is_trained_ = True
        self._totalvi_model.train_indices_ = self.train_set.indices
        self._totalvi_model.test_indices_ = self.test_set.indices
        self._totalvi_model.validation_indices_ = self.validation_set.indices



import scanpy as sc
from scipy import sparse


def remove_sparsity(adata):
    """
        If ``adata.X`` is a sparse matrix, this will convert it in to normal matrix.
        Parameters
        ----------
        adata: :class:`~anndata.AnnData`
            Annotated data matrix.
        Returns
        -------
        adata: :class:`~anndata.AnnData`
            Annotated dataset.
    """
    if sparse.issparse(adata.X):
        new_adata = sc.AnnData(X=adata.X.A, obs=adata.obs.copy(deep=True), var=adata.var.copy(deep=True))
        return new_adata

    return adata
